"""
Functions for running stochastic sampling with Bilby for pre-collected posteriors.

The module provides the `gwpopulation_pipe_analysis` executable.

In order to use many of the other functions you will need a class that provides
various attributes specified in the `gwpopulation_pipe` parser.
"""

#!/usr/bin/env python3

import inspect
import json
import os
import sys
from importlib import import_module

import matplotlib

matplotlib.use("agg")  # noqa

import dill
import numpy as np
import pandas as pd
from bilby.core.sampler import run_sampler
from bilby.core.prior import Constraint, LogUniform, ConditionalPriorDict
from bilby.core.utils import (
    infer_args_from_function_except_n_args,
    logger,
    decode_bilby_json,
)
from bilby_pipe.utils import convert_string_to_dict
from gwpopulation.backend import set_backend
from gwpopulation.conversions import convert_to_beta_parameters
from gwpopulation.hyperpe import HyperparameterLikelihood, RateLikelihood
from gwpopulation.models.mass import (
    BrokenPowerLawPeakSmoothedMassDistribution,
    BrokenPowerLawSmoothedMassDistribution,
    MultiPeakSmoothedMassDistribution,
    SinglePeakSmoothedMassDistribution,
    two_component_primary_mass_ratio,
)
from gwpopulation.models.spin import (
    iid_spin,
    iid_spin_magnitude_beta,
    iid_spin_orientation_gaussian_isotropic,
    independent_spin_magnitude_beta,
    independent_spin_orientation_gaussian_isotropic,
)
from gwpopulation.utils import to_numpy
from scipy.stats import gamma
from tqdm.auto import trange

from . import vt_helper
from .parser import create_parser as create_main_parser
from .utils import (
    get_path_or_local,
    prior_conversion,
    KNOWN_ARGUMENTS,
    MinimumEffectiveSamplesLikelihood,
)


def create_parser():
    parser = create_main_parser()
    parser.add_argument("--prior", help="Prior file readable by bilby.")
    parser.add_argument(
        "--models",
        type=str,
        action="append",
        help="Model functions to evaluate, default is "
        "two component mass and iid spins.",
    )
    parser.add_argument(
        "--vt-models",
        type=str,
        action="append",
        help="Model functions to evaluate for selection, default is no model",
    )
    parser.add_argument(
        "--max-samples",
        default=1e10,
        type=int,
        help="Maximum number of posterior samples per event",
    )
    parser.add_argument(
        "--rate", default=False, type=bool, help="Whether to sample in the merger rate."
    )
    return parser


def load_prior(args):
    filename = get_path_or_local(args.prior_file)
    if filename.endswith(".json"):
        hyper_prior = ConditionalPriorDict.from_json(filename=filename)
    else:
        hyper_prior = ConditionalPriorDict(filename=filename)
    hyper_prior.conversion_function = prior_conversion
    if args.rate:
        hyper_prior["rate"] = LogUniform(
            minimum=1e-1,
            maximum=1e3,
            name="rate",
            latex_label="$R$",
            boundary="reflective",
        )
    return hyper_prior


MODEL_MAP = {
    "two_component_primary_mass_ratio": two_component_primary_mass_ratio,
    "iid_spin": iid_spin,
    "iid_spin_magnitude": iid_spin_magnitude_beta,
    "ind_spin_magnitude": independent_spin_magnitude_beta,
    "iid_spin_orientation": iid_spin_orientation_gaussian_isotropic,
    "two_comp_iid_spin_orientation": iid_spin_orientation_gaussian_isotropic,
    "ind_spin_orientation": independent_spin_orientation_gaussian_isotropic,
    "SmoothedMassDistribution": SinglePeakSmoothedMassDistribution,
    "SinglePeakSmoothedMassDistribution": SinglePeakSmoothedMassDistribution,
    "BrokenPowerLawSmoothedMassDistribution": BrokenPowerLawSmoothedMassDistribution,
    "MultiPeakSmoothedMassDistribution": MultiPeakSmoothedMassDistribution,
    "BrokenPowerLawPeakSmoothedMassDistribution": BrokenPowerLawPeakSmoothedMassDistribution,
}


def _model_class(args):
    if args.cosmo:
        from functools import partial
        from gwpopulation.experimental.cosmo_models import CosmoModel

        cls = partial(CosmoModel, cosmo_model=args.cosmology)
    elif args.backend == "jax":
        from gwpopulation.experimental.jax import NonCachingModel

        cls = NonCachingModel
    else:
        from bilby.hyper.model import Model

        cls = Model
    return cls


def load_model(args):
    if args.models is None:
        args.models = dict(
            mass="two_component_primary_mass_ratio",
            mag="iid_spin_magnitude",
            tilt="iid_spin_orientation",
            redshift="gwpopulation.models.redshift.PowerLawRedshift",
        )
    cls = _model_class(args)
    model = cls([_load_model(model, args) for model in args.models.values()])
    return model


def load_vt(args):
    if args.vt_function == "" or args.vt_file in ["None", None]:
        return vt_helper.dummy_selection
    cls = _model_class(args)
    vt_model = cls([_load_model(model, args) for model in args.vt_models.values()])
    try:
        vt_func = getattr(vt_helper, args.vt_function)
        return vt_func(
            args.vt_file,
            model=vt_model,
            ifar_threshold=args.vt_ifar_threshold,
            snr_threshold=args.vt_snr_threshold,
        )
    except AttributeError:
        return vt_helper.injection_resampling_vt(
            vt_file=args.vt_file,
            model=vt_model,
            ifar_threshold=args.vt_ifar_threshold,
            snr_threshold=args.vt_snr_threshold,
        )


def _load_model(model, args):
    if model[-5:] == ".json":
        model = get_path_or_local(model)
        with open(model, "r") as ff:
            json_model = json.load(ff, object_hook=decode_bilby_json)
        try:
            cls = getattr(import_module(json_model["module"]), json_model["class"])
            _model = cls(**json_model.get("kwargs", dict()))
            logger.info(f"Using {cls} from {json_model['module']}.")
        except KeyError:
            logger.error(f"Failed to load {model} from json file.")
            raise
    elif "." in model:
        module, function = model.rsplit(".", maxsplit=1)
        try:
            _model = getattr(import_module(module), function)
        except ModuleNotFoundError:
            if os.path.exists(f"{module}.py"):
                logger.info(f"Unable to import {module}, using local file")
                sys.path.append(".")
                _model = getattr(import_module(module), function)
            else:
                raise
        logger.info(f"Using {function} from {module}.")
    elif model in MODEL_MAP:
        _model = MODEL_MAP[model]
        logger.info(f"Using {model}.")
    else:
        raise ValueError(f"Model {model} not found.")
    if inspect.isclass(_model):
        if "redshift" in model.lower():
            kwargs = dict(z_max=args.max_redshift)
        elif "mass" in model.lower():
            kwargs = dict(mmin=args.minimum_mass, mmax=args.maximum_mass)
        else:
            kwargs = dict()
        try:
            _model = _model(**kwargs)
            logger.info(f"Created {model} with arguments {kwargs}")
        except TypeError:
            logger.warning(f"Failed to instantiate {model} with arguments {kwargs}")
            _model = _model()
    return _model


def create_likelihood(args, posteriors, model, selection):
    if args.rate:
        if args.enforce_minimum_neffective_per_event:
            raise ValueError(
                "No likelihood available to enforce convergence of Monte Carlo integrals "
                "while sampling over rate."
            )
        likelihood_class = RateLikelihood
    elif args.enforce_minimum_neffective_per_event:
        likelihood_class = MinimumEffectiveSamplesLikelihood
    else:
        likelihood_class = HyperparameterLikelihood
    selection.enforce_convergence = False
    likelihood = likelihood_class(
        posteriors,
        model,
        conversion_function=convert_to_beta_parameters,
        selection_function=selection,
        max_samples=args.max_samples,
        cupy=args.backend == "cupy",
        maximum_uncertainty=args.maximum_uncertainty,
    )

    return likelihood


def get_sampler_kwargs(args):
    sampler_kwargs = dict(nlive=500, nact=2, walks=5)
    if args.sampler_kwargs == "Default":
        sampler_kwargs = dict()
    elif not isinstance(args.sampler_kwargs, dict):
        sampler_kwargs.update(convert_string_to_dict(args.sampler_kwargs))
    else:
        sampler_kwargs = args.sampler_kwargs
    if args.sampler == "cpnest" and "seed" not in sampler_kwargs:
        sampler_kwargs["seed"] = np.random.randint(0, 1e6)
    return sampler_kwargs


def compute_rate_posterior(posterior, selection):
    r"""
    Compute the rate posterior as a post-processing step.

    This method is the same as described in https://dcc.ligo.org/T2000100.
    To get the rate at :math:`z=0` we stop after step four.

    The total surveyed four-volume is given as

    .. math::

        V_{\rm tot}(\Lambda) = T_{\rm obs} \int dz \frac{1}{1+z} \frac{dVc}{dz} \psi(z|\Lambda)

    Note that :math:`\psi(z=0|\Lambda) = 1`

    The sensitive four-volume is then :math:`\mu V_{\rm tot}` where :math:`\mu` is the
    fraction of injections which are found.

    We draw samples from the gamma distribution with mean N_EVENTS + 1

    These samples of this are then divided by the sensitive four-volume to
    give the average rate over the surveyed volume with units :math:`Gpc^{-3}yr^{-1}`.

    Parameters
    ----------
    posterior: pd.DataFrame
        DataFrame containing the posterior samples
    selection: vt_helper.InjectionResamplingVT
        Object that computes:
          - the mean and variance of the survey completeness
          - the total surveyed 4-volume weighted by the redshift distribution
    """
    from numpy import log10
    from .utils import maybe_jit

    if selection == vt_helper.dummy_selection:
        posterior["log_10_rate"] = log10(
            gamma(a=vt_helper.N_EVENTS).rvs(len(posterior))
        )
        return
    else:
        efficiencies = list()
        n_effective = list()
        surveyed_hypervolume = list()
        func = maybe_jit(selection.detection_efficiency)
        for ii in trange(len(posterior), file=sys.stdout):
            parameters = dict(posterior.iloc[ii])
            efficiency, variance = func(parameters)
            efficiencies.append(float(efficiency))
            n_effective.append(float(efficiency**2 / variance))
            surveyed_hypervolume.append(
                float(selection.surveyed_hypervolume(parameters))
            )
        posterior["selection"] = efficiencies
        posterior["pdet_n_effective"] = n_effective
        posterior["surveyed_hypervolume"] = surveyed_hypervolume
        posterior["log_10_rate"] = log10(
            gamma(a=int(vt_helper.N_EVENTS)).rvs(len(posterior))
            / posterior["surveyed_hypervolume"]
            / posterior["selection"]
        )
    posterior["rate"] = 10**posterior.log_10_rate


def fill_minimum_n_effective(posterior, likelihood):
    """
    Compute the minimum per event n effective for each posterior sample.
    This is added to the posterior in place.

    Parameters
    ----------
    posterior: pd.DataFrame
        DataFrame containing posterior distribution
    likelihood: gwpopulation.hyperpe.HyperparameterLikelihood
        The likelihood used in the analysis.

    Returns
    -------

    """
    if not hasattr(likelihood, "per_event_bayes_factors_and_n_effective"):
        logger.info(
            "Likelihood has no method 'per_event_bayes_factors_and_n_effective'"
            " skipping n_effective calculation."
        )
        return
    all_n_effectives = list()
    for ii in trange(len(posterior), file=sys.stdout):
        parameters = dict(posterior.iloc[ii])
        parameters, _ = likelihood.conversion_function(parameters)
        likelihood.parameters.update(parameters)
        likelihood.hyper_prior.parameters.update(parameters)
        _, n_effectives = likelihood.per_event_bayes_factors_and_n_effective()
        all_n_effectives.append(float(min(n_effectives)))
    posterior["min_event_n_effective"] = all_n_effectives
    return


def resample_single_event_posteriors(likelihood, result, save=True):
    """
    Resample the single event posteriors to use the population-informed prior.

    Parameters
    ----------
    likelihood: gwpopulation.hyperpe.HyperparameterLikelihood
        The likelihood object to use.
    result: bilby.core.result.Result
        The result whose posterior should be used for the reweighting.
    save: bool
        Whether to save the samples to file. If `False` the samples will be returned.

    Returns
    -------
    original_samples: dict
        The input samples with the new prior weights in a new `weights` entry.
    reweighted_samples: dict
        The input samples resampled in place according to the new prior weights.
        Note that this will cause samples to be repeated.
    """
    original_samples = likelihood.data
    original_samples["prior"] = likelihood.sampling_prior
    reweighted_samples, weights = likelihood.posterior_predictive_resample(
        result.posterior, return_weights=True
    )
    original_samples["weights"] = weights
    original_samples = {
        key: to_numpy(original_samples[key]) for key in original_samples
    }
    reweighted_samples = {
        key: to_numpy(reweighted_samples[key]) for key in reweighted_samples
    }
    if save:
        with open(f"{result.outdir}/{result.label}_samples.pkl", "wb") as ff:
            dill.dump(
                dict(
                    original=original_samples,
                    reweighted=reweighted_samples,
                    names=result.meta_data["event_ids"],
                    label=result.label,
                ),
                file=ff,
            )
    else:
        return original_samples, reweighted_samples


def generate_extra_statistics(posterior, likelihood):
    from .utils import maybe_jit

    all_samples = list()
    func = maybe_jit(likelihood.generate_extra_statistics)
    for ii in trange(len(posterior), file=sys.stdout):
        parameters = dict(posterior.iloc[ii])
        updated = func(parameters)
        all_samples.append({key: float(updated[key]) for key in updated})
    return pd.DataFrame(all_samples)


def scale_evidences_by_cut(result, likelihood, hyper_prior, N_samples=3000):
    """
    Scale the evidence calculated by dynesty by the fraction of prior samples
    retained after convergence cuts.
    Dynesty calculates the evidence with respect to the prior that has nonzero likelihood support,
    effectively renormalizing the prior. A major cause of zero likelihood support is the convergence
    cuts applied.

    This function scales the evidence by the fraction of prior samples that have nonzero
    likelihood support to undo this renormalization. The field `log_evidence_scaled`
    is added to the result to represent the evidence with respect to the prior supplied
    by the user, whereas `result.log_evidence` is the evidence directly from dynesty,
    which is the evidence with respect to the prior with nonzero likelihood support.

    Parameters
    ----------
    result: bilby.core.result.Result
        The result object containing the evidence to be scaled.
    likelihood: gwpopulation.hyperpe.HyperparameterLikelihood
        The likelihood object used in the analysis.
    hyper_prior: bilby.core.prior.PriorDict
        The prior used in the analysis.
    N_samples: int
        The number of samples to draw from the prior to estimate the fraction of prior samples retained after convergence cuts.

    Returns
    -------
    None

    """
    passed = 0
    for _ in trange(N_samples):
        parameters = dict(hyper_prior.sample())
        parameters, _ = likelihood.conversion_function(parameters)
        likelihood.parameters.update(parameters)
        likelihood.hyper_prior.parameters.update(parameters)
        log_likelihood = likelihood.log_likelihood_ratio()
        if log_likelihood > -np.inf:
            passed += 1

    frac_retained = passed / N_samples
    logger.info(
        f"Fraction of prior samples retained after convergence cuts: {frac_retained}"
    )
    frac_uncertainty = np.sqrt(frac_retained * (1 - frac_retained) / N_samples)
    result.meta_data["log_evidence_scaled"] = result.log_evidence + np.log(
        frac_retained
    )
    result.meta_data["log_bayes_factor_scaled"] = result.log_bayes_factor + np.log(
        frac_retained
    )
    result.meta_data["log_evidence_scaled_err"] = np.sqrt(
        (frac_uncertainty / frac_retained) ** 2
        + (result.log_evidence_err / np.exp(result.log_evidence)) ** 2
    )


def main():
    parser = create_parser()
    args, _ = parser.parse_known_args(sys.argv[1:])

    set_backend(args.backend)

    os.makedirs(args.run_dir, exist_ok=True)
    posterior_file = os.path.join(args.run_dir, "data", f"{args.data_label}.pkl")
    posteriors = pd.read_pickle(posterior_file)
    if not args.cosmo:
        for ii, post in enumerate(posteriors):
            posteriors[ii] = post[post["redshift"] < args.max_redshift]
    vt_helper.N_EVENTS = len(posteriors)
    event_ids = list()
    with open(
        os.path.join(args.run_dir, "data", f"{args.data_label}_posterior_files.txt"),
        "r",
    ) as ff:
        for line in ff.readlines():
            event_ids.append(line.split(":")[0])
    logger.info(f"Loaded {len(posteriors)} posteriors")

    args.models = convert_string_to_dict(
        str(args.models).replace("[", "{").replace("]", "}")
    )
    args.vt_models = convert_string_to_dict(
        str(args.vt_models).replace("[", "{").replace("]", "}")
    )
    hyper_prior = load_prior(args)
    model = load_model(args)
    selection = load_vt(args)

    search_keys = list()
    ignore = ["dataset", "self", "cls"]
    for func in model.models:
        if hasattr(func, "variable_names"):
            param_keys = func.variable_names
        else:
            param_keys = infer_args_from_function_except_n_args(func, n=0)
        param_keys = set(param_keys)
        param_keys.update(KNOWN_ARGUMENTS.get(func, set()))
        for key in param_keys:
            if key in search_keys or key in ignore:
                continue
            search_keys.append(key)
    search_keys.extend(getattr(model, "cosmology_names", list()))

    logger.info(f"Identified keys: {', '.join(search_keys)}")
    for key in list(hyper_prior.keys()):
        if (
            key not in search_keys
            and key != "rate"
            and not isinstance(hyper_prior[key], Constraint)
        ):
            del hyper_prior[key]
        elif (isinstance(hyper_prior[key], Constraint)) and (args.sampler == "numpyro"):
            del hyper_prior[key]

    likelihood = create_likelihood(args, posteriors, model, selection)
    likelihood.parameters.update(hyper_prior.sample())
    likelihood.log_likelihood_ratio()

    if args.injection_file is not None:
        injections = pd.read_json(args.injection_file)
        injection_parameters = dict(injections.iloc[args.injection_index])
    else:
        injection_parameters = None

    if args.backend == "jax" and args.sampler != "numpyro":
        from gwpopulation.experimental.jax import JittedLikelihood

        likelihood = JittedLikelihood(likelihood)
    result = run_sampler(
        likelihood=likelihood,
        priors=hyper_prior,
        label=args.label,
        sampler=args.sampler,
        outdir=os.path.join(args.run_dir, "result"),
        injection_parameters=injection_parameters,
        save="hdf5",
        **get_sampler_kwargs(args),
    )
    result.meta_data["models"] = args.models
    result.meta_data["vt_models"] = args.vt_models
    result.meta_data["event_ids"] = event_ids

    result.posterior = likelihood.conversion_function(result.posterior)[0]

    if args.backend == "jax" and args.sampler == "numpyro":
        from gwpopulation.experimental.jax import JittedLikelihood

        likelihood = JittedLikelihood(likelihood)
    if args.vt_file is not None:
        logger.info("Computing rate posterior")
        compute_rate_posterior(posterior=result.posterior, selection=selection)
    logger.info("Computing n effectives")
    fill_minimum_n_effective(posterior=result.posterior, likelihood=likelihood)
    logger.info("Generating extra statistics")
    result.posterior = generate_extra_statistics(
        posterior=result.posterior, likelihood=likelihood
    )
    logger.info("Scaling evidences by convergence cut")
    scale_evidences_by_cut(result, likelihood, hyper_prior)

    result.save_to_file(extension="hdf5", overwrite=True)

    logger.info("Resampling single event posteriors")
    model = load_model(args)
    selection = load_vt(args)
    likelihood = create_likelihood(args, posteriors, model, selection)
    likelihood.hyper_prior.parameters = likelihood.parameters
    resample_single_event_posteriors(likelihood, result, save=True)



# -*- coding: utf-8 -*-
#
# gwpopulation_pipe documentation build configuration file, created by
# sphinx-quickstart on Fri May 25 12:08:01 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from importlib.metadata import version

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "numpydoc",
    "nbsphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinxarg.ext",
    "sphinx_tabs.tabs",
    "sphinx.ext.viewcode",
]
autosummary_generate = True
add_module_names = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md", ".txt", ".ipynb"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "gwpopulation_pipe"
copyright = "2021-, Colm Talbot"
author = "Colm Talbot"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
fullversion = version("gwpopulation_pipe")

# The short X.Y version.
version = ".".join(fullversion.split(".")[:3])

# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "requirements.txt"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
        "donate.html",
    ]
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "gwpoppipedoc"

numpydoc_show_class_members = False

# nbsphinx options
nbsphinx_execute = "never"



#!/usr/bin/env python3

import glob
import itertools
import os
import shutil
import sys
from pathlib import Path

import htcondor
from bilby.core.sampler import get_sampler_class
from bilby.core.utils import logger
from bilby_pipe.utils import convert_string_to_dict, strip_quotes
from bilby_pipe.job_creation.nodes.analysis_node import touch_checkpoint_files
from htcondor import dags

from .parser import create_parser

MASS_MODELS = dict(
    a="gwpopulation.models.mass.power_law_primary_mass_ratio",
    b="gwpopulation.models.mass.power_law_primary_mass_ratio",
    c="SinglePeakSmoothedMassDistribution",
    d="BrokenPowerLawSmoothedMassDistribution",
    e="MultiPeakSmoothedMassDistribution",
    f="BrokenPowerLawPeakSmoothedMassDistribution",
)

REDSHIFT_MODELS = dict(
    powerlaw="gwpopulation.models.redshift.PowerLawRedshift",
    madaudickinson="gwpopulation.models.redshift.MadauDickinsonRedshift",
)


def relative_topdir(path: str, reference: str) -> str:
    """Returns the top-level directory name of a path relative
    to a reference
    """
    try:
        return str(Path(path).resolve().relative_to(reference))
    except ValueError as exc:
        exc.args = (f"cannot format {path} relative to {reference}",)
        raise


def transfer_container(path: str = None) -> bool:
    return path is not None and os.path.exists(path) and not path.startswith("/cvmfs")


def container_lines(args):
    if args.container is None:
        return dict()
    requirements = "HAS_SINGULARITY=?=True"
    if args.backend == "jax":
        requirements = f"({requirements}) && has_avx"
    config = {
        "MY.SingularityImage": f'"{args.container}"',
        "requirements": requirements,
        "transfer_executable": False,
    }
    return config


def create_base(kind, args, gpus=0):
    log_file = f"{args.log_dir}/population-{kind}-$(label)"
    memory = "8GB"
    if transfer_container(args.container):
        disk = "10GB"
    else:
        disk = "4GB"
    config = {
        "universe": "vanilla",
        "initialdir": os.environ["PWD"],
        "executable": getattr(args, f"{kind}_executable"),
        "output": f"{log_file}.out",
        "error": f"{log_file}.err",
        "log": f"{log_file}.log",
        "request_cpus": "1",
        "request_gpus": gpus,
        "request_memory": memory,
        "request_disk": disk,
        "accounting_group": args.accounting,
        "accounting_group_user": args.user,
        "notification": "error",
        "arguments": _arguments(args, kind).format(
            **{var: f"$({var})" for var in _VARIABLES[kind]}
        ),
        "checkpoint_exit_code": 130,
        "MY.flock_local": True,
    }
    if kind == "collection" or args.pool == "local":
        config["MY.DESIRED_Sites"] = '"none"'
    if args.backend == "jax":
        config["requirements"] = "has_avx"
    if gpus:
        config["require_gpus"] = args.require_gpus
    if kind == "summary":
        config["environment"] = f'CONDA_EXE={os.getenv("CONDA_EXE")}'
    config.update(container_lines(args))
    config.update(transfer_files(args, kind))
    job = htcondor.Submit(config)
    return job


def transfer_files(args, kind):
    run_dir = relative_topdir(args.run_dir, os.environ["PWD"])
    ini_file = f"{run_dir}/{args.label}_config_complete.ini"
    data_dir = f"{run_dir}/data"
    result_dir = f"{run_dir}/result"
    model_files = list()
    for model_list in args.all_models.values():
        for model in model_list:
            if model.endswith(".json"):
                model_files.append(os.path.abspath(model))
    if args.source_files is not None:
        source_files = [os.path.abspath(fname) for fname in args.source_files]
    else:
        source_files = list()
    if kind == "summary":
        inputs = [ini_file]
        inputs.extend(
            [f"{run_dir}/result/{label}_result.hdf5" for label in args.labels]
        )
        outputs = [f"{run_dir}/summary"]
    elif kind == "collection":
        if args.vt_file is None:
            inputs = list()
        elif "*" in args.vt_file:
            inputs = glob.glob(args.vt_file)
        else:
            inputs = [args.vt_file]
        inputs += [ini_file, data_dir]
        outputs = [data_dir]
    elif kind == "analysis":
        inputs = [
            f"{data_dir}/{args.data_label}.pkl",
            f"{data_dir}/{args.data_label}_posterior_files.txt",
            os.path.abspath(args.prior_file),
            ini_file,
        ]
        if args.vt_file is not None:
            inputs.append(f"{data_dir}/injections.pkl")
        cls = get_sampler_class(args.sampler)
        files, dirs = cls.get_expected_outputs(outdir=result_dir, label="$(label)")
        inputs.extend(files)
        inputs.extend(dirs)
        inputs.extend(model_files)
        inputs.extend(source_files)
        outputs = [result_dir]
    elif kind == "format":
        inputs = [
            f"{result_dir}/$(label)_result.hdf5",
            f"{result_dir}/$(label)_samples.pkl",
        ]
        if args.vt_file is not None:
            inputs.append(f"{data_dir}/injections.pkl")
        inputs.extend(model_files)
        inputs.extend(source_files)
        outputs = [f"{result_dir}/$(label)_full_posterior.hdf5"]
    elif kind == "plot":
        inputs = [
            ini_file,
            f"{result_dir}/$(label)_result.hdf5",
            f"{result_dir}/$(label)_samples.pkl",
        ]
        if args.vt_file is not None:
            inputs.append(f"{data_dir}/injections.pkl")
        inputs.extend(model_files)
        inputs.extend(source_files)
        outputs = [result_dir]
    else:
        raise ValueError(f"Unknown job type: {kind}")
    if transfer_container(args.container):
        inputs.append(args.container)
    return dict(
        should_transfer_files="YES",
        transfer_input_files=f"{','.join(inputs)}",
        transfer_output_files=f"{','.join(outputs)}",
        when_to_transfer_output="ON_EXIT_OR_EVICT",
        preserve_relative_paths=True,
        stream_error=True,
        stream_output=True,
    )


def _arguments(args, kind):
    run_dir = relative_topdir(args.run_dir, os.environ["PWD"])
    ini_file = f"{run_dir}/{args.label}_config_complete.ini"
    data_dir = f"{run_dir}/data"
    result_dir = f"{run_dir}/result"
    if args.vt_file is not None:
        vt_file = f"{data_dir}/injections.pkl"
    else:
        vt_file = args.vt_file
    if kind == "analysis":
        arguments = (
            f"{ini_file} --run-dir {run_dir} --label {{label}} "
            f"{{models}} {{vt_models}} --vt-file {vt_file}"
        )
    elif kind == "collection":
        arguments = f"{ini_file} --run-dir {run_dir}"
    elif kind == "format":
        arguments = (
            f"--result-file {result_dir}/{{label}}_result.hdf5 "
            f"--n-samples {args.n_post_samples} --max-redshift {args.max_redshift} "
            f"--minimum-mass {args.minimum_mass} --maximum-mass {args.maximum_mass} "
            f"--injection-file {vt_file} "
            f"--filename {result_dir}/{{label}}_full_posterior.hdf5 "
            f"--samples-file {result_dir}/{{label}}_samples.pkl "
            f"--vt-ifar-threshold {args.vt_ifar_threshold} "
            f"--vt-snr-threshold {args.vt_snr_threshold} "
            f"--backend {args.backend} --cosmology {args.cosmology} "
        )
        if args.cosmo:
            arguments += f"--cosmo "
    elif kind == "plot":
        arguments = (
            f"{ini_file} --run-dir {run_dir} "
            f"--result-file {result_dir}/{{label}}_result.hdf5 "
            f"--samples {result_dir}/{{label}}_samples.pkl "
        )
    elif kind == "summary":
        result_files = " ".join(
            [f"{result_dir}/{label}_result.hdf5" for label in args.labels]
        )
        arguments = (
            f"--config {ini_file} --webdir {run_dir}/summary "
            f"--samples {result_files} --labels {' '.join(args.labels)}"
        )
    return arguments


_VARIABLES = dict(
    analysis=["label", "models", "vt_models"],
    collection=list(),
    format=["label"],
    plot=["label"],
    summary=list(),
)


def check_user(user=None):
    if user is None:
        if "USER" in os.environ:
            user = os.environ.get["USER"]
        else:
            raise ValueError(
                "Argument 'user' must be provided or set in environment variables!"
            )
    return user


def make_submit_files(args):
    subfiles = ["analysis", "collection", "common_format"]
    if args.post_plots:
        subfiles.append("plot")
    if args.make_summary:
        subfiles.append("summary")

    uses_gpu = ["analysis", "common_format", "plot"]

    jobs = dict()
    for label in subfiles:
        if label in uses_gpu:
            gpus = args.request_gpu
        else:
            gpus = 0
        jobs[label] = create_base(label.split("_")[-1], args, gpus)
    return jobs


def update_args(args):
    args.user = check_user(user=args.user)
    args.accounting = "ligo.dev.o4.cbc.bayesianpopulations.parametric"
    args.run_dir = os.path.abspath(args.run_dir)
    args.log_dir = os.path.abspath(args.log_dir)
    args.request_gpu = int(args.request_gpu)

    args.all_models = convert_string_to_dict(args.all_models)
    if args.conda_env is not None:
        env_path = f"{args.conda_env}/bin"
    elif args.container is not None:
        env_path = ""
    else:
        env_path = shutil.which("gwpopulation_pipe").rsplit("/", maxsplit=1)[0]
    if not env_path.endswith("/"):
        env_path += "/"
    args.analysis_executable = f"{env_path}gwpopulation_pipe_analysis"
    args.collection_executable = f"{env_path}gwpopulation_pipe_collection"
    args.format_executable = f"{env_path}gwpopulation_pipe_to_common_format"
    args.plot_executable = f"{env_path}gwpopulation_pipe_plot"
    args.summary_executable = f"{env_path}summarypages"
    args.custom_plotting = (
        os.path.join(os.path.dirname(__file__), "pesummary_plot.py"),
    )
    args.condor_dir = os.path.join(args.run_dir, "submit")
    args.result_dir = os.path.join(args.run_dir, "result")
    args.summary_dir = os.path.join(args.run_dir, "summary")
    args.data_dir = os.path.join(args.run_dir, "data")


def create_directories(args):
    for directory in [
        args.run_dir,
        args.log_dir,
        args.condor_dir,
        args.result_dir,
        args.data_dir,
        args.summary_dir,
    ]:
        if not os.path.isdir(directory):
            os.mkdir(directory)
        elif not os.path.isdir(directory):
            raise IOError(f"{directory} exists and is not a directory.")

    if os.path.isdir(args.existing_data_directory):
        os.rmdir(args.data_dir)
        os.symlink(
            os.path.abspath(args.existing_data_directory),
            os.path.abspath(args.data_dir),
        )
        do_collection = False
    else:
        do_collection = True
    return do_collection


def reduce_name(value):
    if value.endswith(".json"):
        value = value[:-5]
    return value.split(".")[-1]


def write_bash_file(variables, args, bash_file):
    bash_str = "#! /bin/bash\n\n"
    bash_str += f"echo 'Moving to {os.environ['PWD']}'\n"
    bash_str += f"cd {os.environ['PWD']}\n\n"
    if "collection" in variables:
        bash_str += f"{args.collection_executable} {_arguments(args, 'collection')}\n\n"
    for avar, pvar, fvar in zip(
        variables["analysis"], variables["plot"], variables["common_format"]
    ):
        bash_str += f"{args.analysis_executable} {_arguments(args, 'analysis').format(**avar)}\n\n"
        bash_str += (
            f"{args.plot_executable} {_arguments(args, 'plot').format(**pvar)}\n\n"
        )
        bash_str += (
            f"{args.format_executable} {_arguments(args, 'format').format(**fvar)}\n\n"
        )
    if "summary" in variables:
        bash_str += f"{args.summary_executable} {_arguments(args, 'summary')}\n\n"
    bash_str += "cd -\n"
    with open(bash_file, "w") as ff:
        ff.write(bash_str)


def make_dag(args):
    update_args(args)
    do_collection = create_directories(args)

    all_variables = dict()
    if do_collection:
        all_variables["collection"] = [dict(label=args.label)]

    run_dir = relative_topdir(args.run_dir, os.environ["PWD"])

    job_names = list()
    result_files = list()

    all_variables["analysis"] = list()
    all_variables["plot"] = list()
    all_variables["common_format"] = list()

    for values in itertools.product(*args.all_models.values()):
        models = dict(zip(args.all_models.keys(), values))
        prior_name = "_".join(
            [f"{key}_{reduce_name(value)}" for key, value in models.items()]
        )
        job_name = f"{args.label}_{prior_name}"
        job_names.append(job_name)
        result_files.append(f"{run_dir}/result/{job_name}_result.hdf5")

        if models.get("mass", None) in MASS_MODELS:
            models["mass"] = MASS_MODELS[models["mass"]]
        if models.get("redshift", None) in REDSHIFT_MODELS:
            models["redshift"] = REDSHIFT_MODELS[models["redshift"]]
        vt_models = dict()
        for key in models:
            if key in args.vt_parameters:
                vt_models[key] = models[key]
        for key in args.vt_parameters:
            if key not in models:
                logger.error(
                    f"VT parameter {key} not in models. Names in vt-parameters must be in models."
                )
        if len(vt_models) == 0:
            vt_models = {key: models[key] for key in ["mass", "redshift"]}

        models = "--models " + " --models ".join(
            [f"{key}:{value}" for key, value in models.items()]
        )
        vt_models = "--vt-models " + " --vt-models ".join(
            [f"{key}:{value}" for key, value in vt_models.items()]
        )

        all_variables["analysis"].append(
            dict(label=job_name, models=models, vt_models=vt_models)
        )
        touch_checkpoint_files(
            directory=args.result_dir,
            label=job_name,
            sampler=args.sampler,
            result_format="hdf5",
        )

        all_variables["plot"].append(dict(label=job_name))
        all_variables["common_format"].append(dict(label=job_name))

    args.labels = job_names

    if args.make_summary:
        all_variables["summary"] = [dict()]

    def layer_kwargs(kind):
        return dict(
            name=f"{args.condor_dir}/{kind}",
            submit_description=jobs[kind],
            vars=all_variables[kind],
        )

    dag = dags.DAG()
    jobs = make_submit_files(args)
    if do_collection:
        collection_layer = dag.layer(**layer_kwargs("collection"))
        next_layer = collection_layer.child_layer
    else:
        next_layer = dag.layer
    analysis_layer = next_layer(**layer_kwargs("analysis"))
    edges = dict(plot=dags.OneToOne(), common_format=dags.OneToOne(), summary=None)
    for kind in ["plot", "common_format", "summary"]:
        if kind not in all_variables:
            continue
        analysis_layer.child_layer(edge=edges[kind], **layer_kwargs(kind))

    bash_file = f"{args.condor_dir}/{args.label}.sh"
    dag_file = f"{args.condor_dir}/{args.label}.dag"

    dags.write_dag(dag, args.condor_dir, dag_file_name=f"{args.label}.dag")
    write_bash_file(all_variables, args, bash_file)

    print(f"dag file written to {dag_file}")
    print(f"shell script written to {bash_file}")
    print(f"Now run condor_submit_dag {dag_file}")


def main():
    parser = create_parser()
    args, _ = parser.parse_known_args(sys.argv[1:])

    complete_ini_file = f"{args.run_dir}/{args.label}_config_complete.ini"
    make_dag(args)
    parser.write_to_file(
        filename=complete_ini_file,
        args=args,
        overwrite=True,
        include_description=False,
    )
    with open(complete_ini_file, "r") as ff:
        content = ff.readlines()
    for ii, line in enumerate(content):
        content[ii] = strip_quotes(line)
    with open(complete_ini_file, "w") as ff:
        ff.writelines(content)



import inspect
import os

import numpy as np
from bilby.core.sampler.base_sampler import Sampler, signal_wrapper


class NumPyro(Sampler):
    """bilby wrapper for samplers implemented in `NumPyro` (https://pyro.ai/numpyro/)

    All keyword arguments needed for the sampler are passed via the keyword arguments
    passed to this class.
    The needed arguments are automatically determined from the sampler's signature
    other than the :code:`sampler_name` which should be importable from
    :code:`numpyro.infer`.

    Notes
    =====
    This sampler will fail if the likelihood function is not compatible with
    `jax` and `numpyro`.
    """

    sampling_seed_key = "seed"

    default_kwargs = dict(
        sampler_name="NUTS",
        num_warmup=500,
        num_samples=1000,
        num_chains=1,
        verbose=True,
        seed=1234,
    )

    @signal_wrapper
    def run_sampler(self):
        """Runs Nestle sampler with given kwargs and returns the result

        Returns
        =======
        bilby.core.result.Result: Packaged information about the result

        """
        import arviz
        import jax
        import numpyro.infer
        from numpyro.infer import MCMC, Predictive

        rng_key = jax.random.PRNGKey(self.kwargs[self.sampling_seed_key])
        validate_key, sample_key, predictive_key = jax.random.split(rng_key, 3)

        numpyro_model = self.model
        validate_model(numpyro_model, validate_key)

        sampler_class = getattr(numpyro.infer, self.kwargs["sampler_name"])
        sampler = self.evaluate_with_kwargs(sampler_class, numpyro_model)
        mcmc = self.evaluate_with_kwargs(MCMC, sampler)
        mcmc.run(sample_key)

        samples = mcmc.get_samples()
        predictive = Predictive(numpyro_model, samples)(predictive_key)
        predictive = {
            key: predictive[key]
            for key in set(predictive.keys()).difference(samples.keys())
        }
        azdata = arviz.from_numpyro(mcmc, posterior_predictive=predictive)
        keys = [key for key in azdata.posterior.keys() if not key.endswith("_scaled")]

        self.plot_trace(azdata, keys)

        self.result.posterior = azdata.to_dataframe(groups="posterior")[keys]
        self.result.sampler_output = azdata.posterior
        self.result.log_likelihood_evaluations = azdata.log_likelihood[
            "log_likelihood"
        ].values.flatten()
        return self.result

    @property
    def model(self):
        from gwpopulation.experimental.numpyro import construct_numpyro_model

        likelihood_func = self.likelihood_function
        return construct_numpyro_model(
            likelihood=self.likelihood,
            priors=self.priors,
            likelihood_func=likelihood_func,
            **self.filter_kwargs(likelihood_func),
        )

    @property
    def likelihood_function(self):
        from importlib import import_module
        from gwpopulation.experimental.jax import generic_bilby_likelihood_function

        likelihood_func = self.kwargs.get(
            "likelihood_func", generic_bilby_likelihood_function
        )
        if isinstance(likelihood_func, str) and "." not in likelihood_func:
            raise ValueError(
                f"Likelihood function should be importable, got {likelihood_func}"
            )
        elif isinstance(likelihood_func, str):
            split_model = likelihood_func.split(".")
            module = ".".join(split_model[:-1])
            function = split_model[-1]
            likelihood_func = getattr(import_module(module), function)
        return likelihood_func

    def filter_kwargs(self, func):
        return {
            key: self.kwargs[key]
            for key in inspect.signature(func).parameters.keys()
            if key in self.kwargs
        }

    def evaluate_with_kwargs(self, func, *args, **kwargs):
        return func(*args, **kwargs, **self.filter_kwargs(func))

    def plot_trace(self, azdata, keys=None):
        import arviz
        import matplotlib.pyplot as plt

        _ = arviz.plot_trace(azdata, var_names=keys)
        plt.tight_layout()
        plt.savefig(os.path.join(self.outdir, f"{self.label}_trace.png"))
        plt.close()

    def _run_test(self):
        """
        Runs to test whether the sampler is properly running with the given
        kwargs without actually running to the end

        Returns
        =======
        bilby.core.result.Result: Dummy container for sampling results.

        """
        self.kwargs["num_warmup"] = 2
        self.kwargs["num_samples"] = 2
        return self.run_sampler()

    def write_current_state(self):
        """
        Nestle doesn't support checkpointing so no current state will be
        written on interrupt.
        """
        pass


def validate_model(numpyro_model, rng_key):
    from numpyro import handlers
    from numpyro.infer.util import find_valid_initial_params, substitute, trace

    with handlers.seed(rng_seed=0):
        model = numpyro_model
        model_trace = trace(numpyro_model).get_trace()
        (init_params, pe, grad), is_valid = find_valid_initial_params(
            rng_key,
            substitute(
                model,
                data={
                    k: site["value"]
                    for k, site in model_trace.items()
                    if site["type"] in ["plate"]
                },
            ),
        )
        if np.isnan(float(pe)):
            raise ValueError("Initial energy contain NaNs")
        elif any(np.isnan([float(value) for value in grad.values()])):
            raise ValueError("Initial gradient contain NaNs")
        elif not is_valid:
            raise ValueError(f"Invalid initial params {init_params}")



import os
from argparse import ArgumentParser

import dill
import h5py
import numpy as np
import pandas as pd
from bilby.core.result import read_in_result
from bilby.core.utils import logger
from gwpopulation.backend import set_backend, SUPPORTED_BACKENDS
from gwpopulation.utils import to_numpy, xp
from tqdm import tqdm

from .data_analysis import load_model
from .data_collection import evaluate_prior
from .utils import prior_conversion
from .vt_helper import load_injection_data


def create_parser():
    import wcosmo

    parser = ArgumentParser()
    parser.add_argument("-r", "--result-file", help="Bilby result file")
    parser.add_argument(
        "-s", "--samples-file", help="File containing single event samples"
    )
    parser.add_argument("--injection-file", help="File containing injections")
    parser.add_argument("-f", "--filename", default=None, help="Output file name")
    parser.add_argument("--max-redshift", default=2.3, type=float)
    parser.add_argument("--minimum-mass", default=2, type=float)
    parser.add_argument("--maximum-mass", default=100, type=float)
    parser.add_argument(
        "--n-events", type=int, default=None, help="Number of events to draw"
    )
    parser.add_argument(
        "--n-samples",
        type=int,
        default=5000,
        help="The number of population samples to use, default=5000",
    )
    parser.add_argument(
        "--vt-ifar-threshold",
        type=float,
        default=1,
        help="IFAR threshold for resampling injections",
    )
    parser.add_argument(
        "--vt-snr-threshold",
        type=float,
        default=11,
        help="IFAR threshold for resampling injections. "
        "This is only used for O1/O2 injections",
    )
    parser.add_argument(
        "--distance-prior",
        default="euclidean",
        help="Distance prior format, e.g., euclidean, comoving",
    )
    parser.add_argument(
        "--mass-prior",
        default="flat-detector",
        help="Mass prior, only flat-detector is implemented",
    )
    parser.add_argument(
        "--spin-prior",
        default="component",
        help="Spin prior, this should be either component or gaussian",
    )
    parser.add_argument(
        "--backend",
        default="jax",
        choices=SUPPORTED_BACKENDS,
        help="The backend to use, default is jax",
    )
    parser.add_argument(
        "--cosmo", action="store_true", help="Whether to fit cosmological parameters."
    )
    parser.add_argument(
        "--cosmology",
        type=str,
        default="Planck15_LAL",
        help=(
            "Cosmology to use for the analysis, this should be one of "
            f"{', '.join(wcosmo.available.keys())}. Some of these are fixed pre-defined "
            "cosmologies while others are parameterized cosmologies. If a parameterized "
            "cosmology is used the parameters relevant parameters should be included "
            "in the prior specification."
        ),
    )
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()

    set_backend(args.backend)

    result = read_in_result(args.result_file)
    n_samples = min(args.n_samples, len(result.posterior))
    all_samples = dict()
    posterior = result.posterior.sample(n_samples, replace=False)
    posterior = do_spin_conversion(posterior)
    all_samples["posterior"] = posterior

    with open(args.samples_file, "rb") as ff:
        samples = dill.load(ff)
    if "prior" not in samples["original"]:
        data = dict(original=samples["original"])
        evaluate_prior(data, args)
        samples.update(data)
    for key in samples["original"]:
        samples["original"][key] = xp.asarray(samples["original"][key])

    if args.n_events:
        n_draws = args.n_events
        logger.info(f"Number of draws set to {n_draws}.")
    else:
        n_draws = len(samples["names"])
        logger.info(f"Number of draws equals number of events, {n_draws}.")

    logger.info("Generating observed populations.")
    args.models = result.meta_data["models"]
    model = load_model(args)
    observed_dataset = resample_events_per_population_sample(
        posterior=posterior,
        samples=samples["original"],
        model=model,
        n_draws=len(samples["names"]),
    )

    for ii, name in enumerate(samples["names"]):
        new_posterior = pd.DataFrame()
        for key in observed_dataset:
            new_posterior[f"{name}_{key}"] = to_numpy(observed_dataset[key][:, ii])
        all_samples[name] = new_posterior

    if args.injection_file is None:
        logger.info("Can't generate predicted populations without VT file.")
    else:
        logger.info("Generating predicted populations.")
        args.models = result.meta_data["vt_models"]
        vt_data = load_injection_data(
            args.injection_file,
            ifar_threshold=args.vt_ifar_threshold,
            snr_threshold=args.vt_snr_threshold,
        ).to_dict()
        model = load_model(args)
        synthetic_dataset = resample_injections_per_population_sample(
            posterior=posterior,
            data=vt_data,
            model=model,
            n_draws=n_draws,
        )

        for ii in range(n_draws):
            new_posterior = pd.DataFrame()
            for key in synthetic_dataset:
                new_posterior[f"synthetic_{key}_{ii}"] = to_numpy(
                    synthetic_dataset[key][:, ii]
                )
            all_samples[f"synthetic_{ii}"] = new_posterior

    if args.filename is None:
        filename = os.path.join(result.outdir, f"{result.label}_full_posterior.hdf5")
    else:
        filename = args.filename
    save_to_common_format(
        posterior=all_samples, events=samples["names"], filename=filename
    )


def do_spin_conversion(posterior):
    """Utility function to convert between beta distribution parameterizations."""
    original_keys = list(posterior.keys())
    posterior = prior_conversion(posterior)
    for key in ["amax", "amax_1", "amax_2"]:
        if key not in original_keys and key in posterior:
            del posterior[key]
    return posterior


def save_to_common_format(posterior, events, filename):
    """
    Save the data to the common hdf5 format.

    Parameters
    ----------
    posterior: [np.ndarray, pd.DataFrame]
        The posterior to be saved.
    events: list
        The names of each of the events in the dataset.
    filename: str
        the output filename

    """
    for key in posterior:
        if isinstance(posterior[key], pd.DataFrame):
            posterior[key] = data_frame_to_sarray(posterior[key])
    events = [event.encode() for event in events]
    logger.info(f"Writing data to {filename}")
    with h5py.File(filename, "w") as ff:
        for key in posterior:
            ff[f"samples/{key}"] = posterior[key]
        ff["events"] = events


def read_common_format(filename, data_format="numpy"):
    """
    Read a posterior file in the common format

    Parameters
    ----------
    filename: str
        The path to the file to read.
    data_format: str
        The format to return the data in, can be either `numpy` or `pandas`.

    Returns
    -------
    output: [np.ndarray, pd.DataFrame]
        The posterior in the requested format.
    events: list
        The event names.
    """
    with h5py.File(filename, "r") as ff:
        data = {key: ff["samples"][key][:] for key in ff["samples"]}
        events = [event.decode() for event in ff["events"]]
    if data_format == "numpy":
        return data, events
    elif data_format == "pandas":
        output = dict()
        for key in data:
            data_frame = pd.DataFrame()
            for name in data[key].dtype.names:
                data_frame[name] = data[key][name]
            output[key] = data_frame
        return output, events
    else:
        raise ValueError(f"Data format {data_format} not implemented.")


def resample_events_per_population_sample(posterior, samples, model, n_draws):
    """
    Resample the input posteriors with a fiducial prior to the population
    informed distribution. This returns a single sample for each event for
    each passed hyperparameter sample.

    See, e.g., section IIIC of `Moore and Gerosa <https://arxiv.org/abs/2108.02462>`_
    for a description of the method.

    Parameters
    ----------
    posterior: pd.DataFrame
        Hyper-parameter samples to use for the reweighting.
    samples: dict
        Posterior samples with the fiducial prior.
    model: bilby.hyper.model.Model
        Object that implements a `prob` method that will calculate the population
        probability.
    n_draws: int
        The number of samples to draw. This should generally be the number of events.
        This will return one sample per input event.

    Returns
    -------
    observed_dataset: dict
        The observed dataset of the events with the population informed prior.
    """
    from .utils import maybe_jit

    @maybe_jit
    def extract_choices(parameters, seed=None):
        model.parameters.update(parameters)
        if hasattr(model, "cosmology_names"):
            data, jacobian = model.detector_frame_to_source_frame(samples, **parameters)
        else:
            data = samples
            jacobian = 1
        weights = model.prob(data) / data["prior"] / jacobian
        weights = (weights.T / xp.sum(weights, axis=-1)).T
        if "jax" in xp.__name__:
            seeds = random.split(random.PRNGKey(seed), len(weights))
            choices = [
                random.choice(seeds[ii], len(weights[ii]), p=weights[ii])
                for ii in range(n_draws)
            ]
        else:
            choices = [
                random.choice(
                    len(weights[ii]),
                    p=weights[ii],
                    size=1,
                ).squeeze()
                for ii in range(n_draws)
            ]
        return xp.asarray(choices)

    points = posterior.to_dict(orient="records")
    if "jax" in xp.__name__:
        from jax import random

        seeds = np.random.choice(100000, len(posterior))
    else:
        random = xp.random
        seeds = [None] * len(posterior)
    inputs = list(zip(points, seeds))
    all_choices = xp.asarray(
        [extract_choices(point, seed) for point, seed in tqdm(inputs)]
    )

    @maybe_jit
    def resample(samples, choices):
        return xp.asarray([samples[ii, choice] for ii, choice in enumerate(choices)])

    observed_dataset = dict()
    for key in samples:
        observed_dataset[key] = to_numpy(
            xp.asarray([resample(samples[key], choices) for choices in all_choices])
        )
    return observed_dataset


def resample_injections_per_population_sample(posterior, data, model, n_draws):
    """
    Resample the input data with a fiducial prior to the population
    informed distribution. This returns a fixed number of samples for
    each passed hyperparameter sample.

    This is designed to be used with found injections.

    See, e.g., section IIIC of `Moore and Gerosa <https://arxiv.org/abs/2108.02462>`_
    for a description of the method.

    Parameters
    ----------
    posterior: pd.DataFrame
        Hyper-parameter samples to use for the reweighting.
    data: dict
        Input data samples to be reweighted.
    model: bilby.hyper.model.Model
        Object that implements a `prob` method that will calculate the population
        probability.
    n_draws: int
        The number of samples to draw. This should generally be the number of events.
        This will return one sample per input event.

    Returns
    -------
    observed_dataset: dict
        The observed dataset of the input data with the population informed prior.
    """
    from .utils import maybe_jit

    @maybe_jit
    def extract_choices(parameters, seed):
        model.parameters.update(parameters)
        if hasattr(model, "cosmology_names"):
            samples, jacobian = model.detector_frame_to_source_frame(data, **parameters)
        else:
            samples = data
            jacobian = 1
        weights = model.prob(samples) / samples["prior"] / jacobian
        weights /= xp.sum(weights)
        if "jax" in xp.__name__:
            choices = choice(
                PRNGKey(seed),
                len(weights),
                p=weights,
                shape=(n_draws,),
                replace=False,
            )
        else:
            weights = to_numpy(weights)
            choices = choice(len(weights), p=weights, size=n_draws, replace=False)
        return choices

    if "jax" in xp.__name__:
        from jax.random import PRNGKey, choice

        seeds = np.random.choice(100000, len(posterior))
    else:
        # FIXME: cupy.random.choice doesn't support p != None and replace=False
        # cupy==13.0.0 (240202)
        from numpy.random import choice

        seeds = [None] * len(posterior)
    points = posterior.to_dict(orient="records")
    inputs = list(zip(points, seeds))
    all_choices = to_numpy(
        xp.asarray([extract_choices(point, seed) for point, seed in tqdm(inputs)])
    )

    @maybe_jit
    def resample(data, choices):
        """
        For some reason even this simple function significantly benefits from jit
        """
        return data[choices]

    synthetic_dataset = dict()
    for key in data:
        if not isinstance(data[key], xp.ndarray):
            continue
        synthetic_dataset[key] = xp.asarray(
            [resample(data[key], choices) for choices in all_choices]
        )
    return synthetic_dataset


def data_frame_to_sarray(data_frame):
    """
    Convert a pandas DataFrame object to a numpy structured array.
    This is functionally equivalent to but more efficient than
    `np.array(df.to_array())`.

    lifted from https://stackoverflow.com/questions/30773073/save-pandas-dataframe-using-h5py-for-interoperabilty-with-other-hdf5-readers

    Parameters
    -----------
    data_frame: pd.DataFrame
        the data frame to convert

    Returns
    -------
    output: np.ndarray
        a numpy structured array representation of df
    """

    types = [
        (data_frame.columns[index], data_frame[key].dtype.type)
        for (index, key) in enumerate(data_frame.columns)
    ]
    dtype = np.dtype(types)
    output = np.zeros(data_frame.values.shape[0], dtype)
    for (index, key) in enumerate(output.dtype.names):
        output[key] = data_frame.values[:, index]
    return output



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from bilby.hyper.model import Model
from gwpopulation.conversions import convert_to_beta_parameters
from gwpopulation.utils import to_numpy, xp
from gwpopulation.models.mass import (
    double_power_law_primary_power_law_mass_ratio,
    two_component_primary_mass_ratio,
)
from gwpopulation.models.spin import (
    independent_spin_magnitude_beta,
    independent_spin_orientation_gaussian_isotropic,
)
from tqdm import tqdm

MAX_SAMPLES = 10000

__single_plots__ = [
    "mass_spectrum_plot",
    "spin_magnitude_spectrum_plot",
    "spin_orientation_spectrum_plot",
]
__comparison_plots__ = [
    "mass_spectrum_plot",
    "spin_magnitude_spectrum_plot",
    "spin_orientation_spectrum_plot",
]


def mass_spectrum_plot(parameters, samples, labels=None, limits=None, rate=False):
    if limits is None:
        limits = [5, 95]

    if labels is not None and len(labels) > 1:
        all_samples = [
            pd.DataFrame(samps, columns=params)
            for samps, params in zip(samples, parameters)
        ]
    else:
        all_samples = [pd.DataFrame(samples, columns=parameters)]
        labels = [labels]
    mass_1 = xp.linspace(2, 100, 1000)
    mass_ratio = xp.linspace(0.1, 1, 500)
    mass_1_grid, mass_ratio_grid = xp.meshgrid(mass_1, mass_ratio)

    fig, axs = plt.subplots(2, 1, figsize=(12, 8))

    peak_1 = 0
    peak_2 = 0

    for jj, samples in enumerate(all_samples):
        data = dict(mass_1=mass_1_grid, mass_ratio=mass_ratio_grid)
        lines = dict(mass_1=list(), mass_ratio=list())
        ppd = xp.zeros_like(data["mass_1"])

        if len(samples) > MAX_SAMPLES:
            samples = samples.sample(MAX_SAMPLES)
        else:
            samples = samples

        for func in [
            two_component_primary_mass_ratio,
            double_power_law_primary_power_law_mass_ratio,
        ]:
            model = Model([func])
            try:
                model.parameters.update(dict(samples.iloc[0]))
                model.prob(dict(mass_1=xp.asarray([30]), mass_ratio=xp.asarray([0.9])))
                break
            except:
                pass

        for ii in tqdm(range(len(samples))):
            parameters = dict(samples.iloc[ii])
            model.parameters.update(parameters)
            prob = model.prob(data)
            if rate:
                prob *= parameters["rate"]
            ppd += prob

            mass_1_prob = xp.trapz(prob, mass_ratio, axis=0)
            mass_ratio_prob = xp.trapz(prob, mass_1, axis=-1)

            lines["mass_1"].append(mass_1_prob)
            lines["mass_ratio"].append(mass_ratio_prob)
        for key in lines:
            lines[key] = np.vstack([to_numpy(line) for line in lines[key]])

        ppd /= len(samples)
        ppd = to_numpy(ppd)

        mass_1 = to_numpy(mass_1)
        mass_ratio = to_numpy(mass_ratio)

        mass_1_ppd = np.trapz(ppd, mass_ratio, axis=0)
        mass_ratio_ppd = np.trapz(ppd, mass_1, axis=-1)

        label = labels[jj]
        if label is not None:
            label = label.replace("_", " ")

        axs[0].semilogy(mass_1, mass_1_ppd, label=label)
        axs[0].fill_between(
            mass_1,
            np.percentile(lines["mass_1"], limits[0], axis=0),
            np.percentile(lines["mass_1"], limits[1], axis=0),
            alpha=0.5,
        )
        _peak_1 = max(np.percentile(lines["mass_1"], limits[1], axis=0))
        peak_1 = max(peak_1, _peak_1)
        axs[1].semilogy(mass_ratio, mass_ratio_ppd)
        axs[1].fill_between(
            mass_ratio,
            np.percentile(lines["mass_ratio"], limits[0], axis=0),
            np.percentile(lines["mass_ratio"], limits[1], axis=0),
            alpha=0.5,
        )
        _peak_2 = max(np.percentile(lines["mass_ratio"], limits[1], axis=0))
        peak_2 = max(peak_2, _peak_2)

    axs[0].set_xlim(2, 60)
    axs[0].set_ylim(peak_1 / 1000, peak_1 * 1.1)
    axs[0].set_xlabel("$m_{1}$ [$M_{\\odot}$]")
    if rate:
        ylabel = "$\\frac{dN}{dm_{1}}$"
    else:
        ylabel = "$p(m_{1})$ [$M_{\\odot}^{-1}$]"
    axs[0].set_ylabel(ylabel)
    if len(labels) > 1:
        axs[0].legend(loc="best")

    axs[1].set_xlim(0.1, 1)
    axs[1].set_ylim(peak_2 / 10000, peak_2 * 1.1)
    axs[1].set_xlabel("$q$")
    if rate:
        ylabel = "$\\frac{dN}{dq}$"
    else:
        ylabel = "$p(q)$"
    axs[1].set_ylabel(ylabel)

    plt.tight_layout()
    return fig


def spin_magnitude_spectrum_plot(
    parameters, samples, labels=None, limits=None, rate=False
):
    if limits is None:
        limits = [5, 95]

    if labels is not None and len(labels) > 1:
        all_samples = [
            pd.DataFrame(samps, columns=params)
            for samps, params in zip(samples, parameters)
        ]
    else:
        all_samples = [pd.DataFrame(samples, columns=parameters)]
        labels = [labels]
    mags = xp.linspace(0, 1, 1000)
    a_1, a_2 = xp.meshgrid(mags, mags)
    model = Model([independent_spin_magnitude_beta])

    fig, axs = plt.subplots(1, 2, figsize=(12, 4))

    for jj, samples in enumerate(all_samples):
        for key in ["amax", "mu_chi", "sigma_chi"]:
            if f"{key}_1" not in samples:
                samples[f"{key}_1"] = samples[key]
                samples[f"{key}_2"] = samples[key]
                del samples[key]

        data = dict(a_1=a_1, a_2=a_2)
        lines = dict(a_1=list(), a_2=list())
        ppd = xp.zeros_like(data["a_1"])

        samples = convert_to_beta_parameters(samples)[0]

        if len(samples) > MAX_SAMPLES:
            samples = samples.sample(MAX_SAMPLES)
        else:
            samples = samples

        for ii in tqdm(range(len(samples))):
            parameters = dict(samples.iloc[ii])
            model.parameters.update(parameters)
            prob = model.prob(data)
            if rate:
                prob *= parameters["rate"]
            ppd += prob

            a_1_prob = xp.trapz(prob, mags, axis=0)
            a_2_prob = xp.trapz(prob, mags, axis=-1)

            lines["a_1"].append(a_1_prob)
            lines["a_2"].append(a_2_prob)
        for key in lines:
            lines[key] = np.vstack([to_numpy(line) for line in lines[key]])

        ppd /= len(samples)
        ppd = to_numpy(ppd)

        mags = to_numpy(mags)

        a_1_ppd = np.trapz(ppd, mags, axis=0)
        a_2_ppd = np.trapz(ppd, mags, axis=1)

        label = labels[jj]
        if label is not None:
            label = label.replace("_", " ")

        axs[0].plot(mags, a_1_ppd, label=label)
        axs[0].fill_between(
            mags,
            np.percentile(lines["a_1"], limits[0], axis=0),
            np.percentile(lines["a_1"], limits[1], axis=0),
            alpha=0.5,
        )
        axs[1].plot(mags, a_2_ppd)
        axs[1].fill_between(
            mags,
            np.percentile(lines["a_2"], limits[0], axis=0),
            np.percentile(lines["a_2"], limits[1], axis=0),
            alpha=0.5,
        )

    for ii in [1, 2]:
        axs[ii - 1].set_xlim(0, 1)
        axs[ii - 1].set_ylim(0)
        axs[ii - 1].set_xlabel(f"$a_{ii}$")
        if rate:
            ylabel = f"$\\frac{{dN}}{{da_{ii}}}$"
        else:
            ylabel = f"$p(a_{ii})$"
        axs[ii - 1].set_ylabel(ylabel)
    if len(labels) > 1:
        axs[0].legend(loc="best")

    plt.tight_layout()
    return fig


def spin_orientation_spectrum_plot(
    parameters, samples, labels=None, limits=None, rate=False
):
    if limits is None:
        limits = [5, 95]

    if labels is not None and len(labels) > 1:
        all_samples = [
            pd.DataFrame(samps, columns=params)
            for samps, params in zip(samples, parameters)
        ]
    else:
        all_samples = [pd.DataFrame(samples, columns=parameters)]
        labels = [labels]
    cos_tilts = xp.linspace(-1, 1, 1000)
    cos_tilt_1, cos_tilt_2 = xp.meshgrid(cos_tilts, cos_tilts)

    model = Model([independent_spin_orientation_gaussian_isotropic])

    fig, axs = plt.subplots(1, 2, figsize=(12, 4))

    for jj, samples in enumerate(all_samples):
        if "sigma_1" not in samples:
            samples["sigma_1"] = samples["sigma_spin"]
            samples["sigma_2"] = samples["sigma_spin"]

        data = dict(cos_tilt_1=cos_tilt_1, cos_tilt_2=cos_tilt_2)
        lines = dict(cos_tilt_1=list(), cos_tilt_2=list())
        ppd = xp.zeros_like(data["cos_tilt_1"])

        if len(samples) > MAX_SAMPLES:
            samples = samples.sample(MAX_SAMPLES)
        else:
            samples = samples

        for ii in tqdm(range(len(samples))):
            parameters = dict(samples.iloc[ii])
            model.parameters.update(parameters)
            prob = model.prob(data)
            if rate:
                prob *= parameters["rate"]
            ppd += prob

            cos_tilt_1_prob = xp.trapz(prob, cos_tilts, axis=0)
            cos_tilt_2_prob = xp.trapz(prob, cos_tilts, axis=1)

            lines["cos_tilt_1"].append(cos_tilt_1_prob)
            lines["cos_tilt_2"].append(cos_tilt_2_prob)
        for key in lines:
            lines[key] = np.vstack([to_numpy(line) for line in lines[key]])

        ppd /= len(samples)
        ppd = to_numpy(ppd)

        cos_tilts = to_numpy(cos_tilts)

        cos_tilt_1_ppd = np.trapz(ppd, cos_tilts, axis=0)
        cos_tilt_2_ppd = np.trapz(ppd, cos_tilts, axis=1)

        label = labels[jj]
        if label is not None:
            label = label.replace("_", " ")

        axs[0].plot(cos_tilts, cos_tilt_1_ppd, label=label)
        axs[0].fill_between(
            cos_tilts,
            np.percentile(lines["cos_tilt_1"], limits[0], axis=0),
            np.percentile(lines["cos_tilt_1"], limits[1], axis=0),
            alpha=0.5,
        )
        axs[1].plot(cos_tilts, cos_tilt_2_ppd)
        axs[1].fill_between(
            cos_tilts,
            np.percentile(lines["cos_tilt_2"], limits[0], axis=0),
            np.percentile(lines["cos_tilt_2"], limits[1], axis=0),
            alpha=0.5,
        )

    if rate:
        ylabel = "$\\frac{{dN}}{{d\\cos t_{}}}$"
    else:
        ylabel = "$p(\\cos t_{})$"
    for ii in [1, 2]:
        axs[ii - 1].set_xlim(0, 1)
        axs[ii - 1].set_ylim(0)
        axs[ii - 1].set_xlabel(f"$\\cos t_{ii}$")
        axs[ii - 1].set_ylabel(ylabel.format(str(ii)))
    if len(labels) > 1:
        axs[0].legend(loc="best")

    plt.tight_layout()
    return fig



import os
import shutil
from argparse import ArgumentParser

from bilby.core.utils import logger
from bilby_pipe.parser import StoreBoolean

from .create_injections import create_injections
from .main import check_user

TEMPLATE = """################################################################################
## Generic arguments
################################################################################

label=pp_test_{idx}
run-dir=.
log-dir=logs/{idx}
user={user}
prior-dir=prior_files
vt-model=""
request-gpu={gpu}

################################################################################
## Analysis models
################################################################################

mass-models=[{mass}]
magnitude-models=[{magnitude}]
tilt-models=[{orientation}]
redshift-models=[dummy]

################################################################################
## Data collection arguments
################################################################################

data-label=pp_test_{idx}
injection-file={injection_file}
injection-index={idx}
n-simulations={n_simulation}
sample-from-prior={sample_from_prior}
plot=False

################################################################################
## Post arguments
################################################################################

post-plots=False
make-summary=False
"""


def setup_pp_test():
    parser = ArgumentParser()
    parser.add_argument(
        "--run-dir", default="pp_test", help="Directory to run the test in."
    )
    parser.add_argument("--prior-file", help="Prior file to generate samples from.")
    parser.add_argument(
        "--user", help="albert.einstein style user name.", default=None, type=str
    )
    parser.add_argument(
        "--n-simulation", type=int, default=100, help="Number of jobs to run."
    )
    parser.add_argument(
        "--n-events", type=int, default=40, help="Number of events to simulate per run."
    )
    parser.add_argument(
        "--sample-from-prior",
        action=StoreBoolean,
        help="Simulate posteriors from prior.",
    )
    parser.add_argument(
        "--mass-model",
        choices=["a", "b", "c", "d"],
        default="c",
        help="Mass model to use.",
    )
    parser.add_argument(
        "--magnitude-model",
        choices=["iid", "ind"],
        default="ind",
        help="Spin magnitude model to use.",
    )
    parser.add_argument(
        "--orientation-model",
        choices=["iid", "ind"],
        default="ind",
        help="Spin orientation model to use.",
    )
    parser.add_argument(
        "--submit",
        default=False,
        action=StoreBoolean,
        help="Whether to submit to condor.",
    )
    parser.add_argument(
        "--request-gpu",
        default=False,
        action=StoreBoolean,
        help="Whether to use a GPU for the relevant jobs.",
    )
    args = parser.parse_args()
    args.user = check_user(user=args.user)

    if not os.path.exists(args.run_dir):
        os.mkdir(args.run_dir)

    samples = create_injections(
        prior_file=args.prior_file, n_simulation=args.n_simulation
    )
    injection_file = os.path.abspath(
        os.path.join(args.run_dir, "injection_values.json")
    )
    samples.to_json(injection_file)

    bash_script = f"cd {args.run_dir}\n"
    total_dag = ""
    for ii in range(args.n_simulation):
        config_text = TEMPLATE.format(
            injection_file=injection_file,
            idx=ii,
            user=args.user,
            run_dir=args.run_dir,
            n_simulation=args.n_events,
            gpu=args.request_gpu,
            sample_from_prior=args.sample_from_prior,
            mass=args.mass_model,
            magnitude=args.magnitude_model,
            orientation=args.orientation_model,
        )
        config_filename = os.path.join(args.run_dir, f"config_{ii}.ini")
        with open(config_filename, "w") as ff:
            ff.write(config_text)
        bash_script += f"gwpopulation_pipe config_{ii}.ini\n"
        temp_dag_file = os.path.abspath(
            os.path.join(args.run_dir, "submit", f"pp_test_{ii}.dag")
        )
        total_dag += f"SUBDAG EXTERNAL pp_test_{ii} {temp_dag_file} DIR .\n"
        total_dag += f"RETRY pp_test_{ii} 0\n"

    bash_script += "cd -\n"

    multidag_file = os.path.join(os.path.abspath(args.run_dir), "multidag.dag")
    run_file = os.path.join(os.path.abspath(args.run_dir), "run.sh")

    bash_script += f"condor_submit_dag {multidag_file}\n"

    with open(multidag_file, "w") as ff:
        ff.write(total_dag)

    with open(run_file, "w") as ff:
        ff.write(bash_script)

    with open(os.path.join(args.run_dir, "post.sh"), "w") as ff:
        ff.write(f"bilby_pipe_pp_test .")

    for sub_dir in ["prior_files", "logs"]:
        if not os.path.exists(os.path.join(args.run_dir, sub_dir)):
            os.mkdir(os.path.join(args.run_dir, sub_dir))

    shutil.copy(args.prior_file, os.path.join(args.run_dir, "prior_files"))

    command = f"bash {run_file}"
    if args.submit:
        os.system(command=command)
    else:
        logger.info(f"Now run\n$ {command}")


def main():
    pass



from dataclasses import dataclass, field

import argparse
import glob
from functools import reduce

import dill
import h5py
import numpy as np
from bilby.core.utils import logger

from gwpopulation.utils import xp
from gwpopulation.vt import GridVT, ResamplingVT
from .utils import (
    get_cosmology,
    get_path_or_local,
    chi_eff_from_components,
    chi_p_from_components,
)
from .analytic_spin_prior import prior_chieff_chip_isotropic

N_EVENTS = np.nan


def dummy_selection(*args, **kwargs):
    return 1


def mass_only_scalar_calibrated_grid_vt(vt_file, model):
    selection = _mass_only_calibrated_grid_vt(
        vt_file=vt_file,
        model=model,
        calibration="scalar",
    )
    return selection


def mass_only_linear_calibrated_grid_vt(vt_file, model):
    selection = _mass_only_calibrated_grid_vt(
        vt_file=vt_file,
        model=model,
        calibration="linear",
    )
    return selection


def mass_only_quadratic_calibrated_grid_vt(vt_file, model):
    selection = _mass_only_calibrated_grid_vt(
        vt_file=vt_file,
        model=model,
        calibration="quadratic",
    )
    return selection


gaussian_mass_only_scalar_calibrated_grid_vt = mass_only_scalar_calibrated_grid_vt
gaussian_mass_only_linear_calibrated_grid_vt = mass_only_linear_calibrated_grid_vt
gaussian_mass_only_quadratic_calibrated_grid_vt = mass_only_quadratic_calibrated_grid_vt
broken_mass_only_scalar_calibrated_grid_vt = mass_only_scalar_calibrated_grid_vt
broken_mass_only_linear_calibrated_grid_vt = mass_only_linear_calibrated_grid_vt
broken_mass_only_quadratic_calibrated_grid_vt = mass_only_quadratic_calibrated_grid_vt


def _mass_only_calibrated_grid_vt(vt_file, model, calibration=None):
    import h5py

    model = model
    vt_data = dict()
    with h5py.File(vt_file) as _vt_data:
        vt_data["vt"] = _vt_data["vt_early_high"][:]
        if calibration is not None:
            vt_data["vt"] *= _vt_data[f"{calibration}_calibration"][:]
        vt_data["vt"] = xp.asarray(vt_data["vt"])
        vt_data["mass_1"] = xp.asarray(_vt_data["m1"][:])
        vt_data["mass_ratio"] = xp.asarray(_vt_data["q"][:])

    selection = GridVT(model=model, data=vt_data)
    return selection


@dataclass
class VTData:
    prior: np.ndarray
    total_generated: int
    analysis_time: float
    mass_1: np.ndarray = field(default=None)
    mass_1_detector: np.ndarray = field(default=None)
    mass_ratio: np.ndarray = field(default=None)
    a_1: np.ndarray = field(default=None)
    a_2: np.ndarray = field(default=None)
    cos_tilt_1: np.ndarray = field(default=None)
    cos_tilt_2: np.ndarray = field(default=None)
    chi_eff: np.ndarray = field(default=None)
    chi_p: np.ndarray = field(default=None)
    redshift: np.ndarray = field(default=None)
    luminosity_distance: np.ndarray = field(default=None)
    mass_2: np.ndarray = field(default=None)
    mass_2_detector: np.ndarray = field(default=None)

    def append(self, other):
        self_sample_rate = self.analysis_time / self.total_generated
        other_sample_rate = other.analysis_time / other.total_generated
        self_weight = 2 * self_sample_rate / (self_sample_rate + other_sample_rate)
        other_weight = 2 * other_sample_rate / (self_sample_rate + other_sample_rate)
        for key in self.__dataclass_fields__:
            value = getattr(self, key)
            alt = getattr(other, key)
            if key == "mass_2":
                if value is None and alt is None:
                    continue
                elif value is not None and alt is not None:
                    setattr(self, key, np.concatenate([value, alt]))
                else:
                    raise ValueError("mass_2 is only defined for one VTData object")
            elif key in ["total_generated", "analysis_time"]:
                setattr(self, key, value + alt)
            elif key == "prior":
                setattr(
                    self,
                    key,
                    np.concatenate([value * self_weight, alt * other_weight]),
                )
            else:
                setattr(self, key, np.concatenate([value, alt]))

    def __add__(self, other):
        new = VTData(**self.__dict__)
        new += other
        return new

    def __iadd__(self, other):
        self.append(other)
        return self

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def get(self, key, alt):
        return self.__dict__.get(key, alt)

    def to_dict(self, keys=None):
        if keys is None:
            keys = list(self.__dict__.keys())
        data = {key: getattr(self, key) for key in keys}
        return {key: value for key, value in data.items() if value is not None}


def load_injection_data(vt_file, ifar_threshold=1, snr_threshold=11):
    """
    Load the injection file in the O3 injection file format.

    For mixture files and multiple observing run files we only
    have the full `sampling_pdf`.

    We use a different parameterization than the default so we require a few
    changes.

    - we parameterize the model in terms of primary mass and mass ratio and
      the injections are generated in primary and secondary mass. The Jacobian
      is `primary mass`.
    - we parameterize spins in spherical coordinates, neglecting azimuthal
      parameters. The injections are parameterized in terms of cartesian
      spins. The Jacobian is `1 / (2 pi magnitude ** 2)`.

    For O3 injections we threshold on FAR.
    For O1/O2 injections we threshold on SNR as there is no FAR
    provided by the search pipelines.

    Parameters
    ----------
    vt_file: str
        The path to the hdf5 file containing the injections.
    ifar_threshold: float
        The threshold on inverse false alarm rate in years. Default=1.
    snr_threshold: float
        The SNR threshold when there is no FAR. Default=11.

    Returns
    -------
    gwpop_data: dict
        Data required for evaluating the selection function.

    """
    logger.info(f"Loading VT data from {vt_file}.")
    if vt_file.endswith(".pkl"):
        with open(vt_file, "rb") as ff:
            data = dill.load(ff)
        for key, value in data.items():
            if isinstance(value, np.ndarray):
                data[key] = xp.asarray(value)
        return VTData(**data)

    with h5py.File(vt_file, "r") as ff:
        if "injections" in ff:
            data = ff["injections"]
            total_generated = int(data.attrs["total_generated"][()])
            analysis_time = data.attrs["analysis_time_s"][()] / 365.25 / 24 / 60 / 60
        elif "events" in ff:
            keys_of_interest = {
                "mass1_source",
                "mass2_source",
                "mass_1_source",
                "mass_2_source",
                "spin1x",
                "spin1y",
                "spin1z",
                "spin2x",
                "spin2y",
                "spin2z",
                "redshift",
                "z",
                "sampling_pdf",
                "lnpdraw_mass1_source_mass2_source_redshift_spin1x_spin1y_spin1z_spin2x_spin2y_spin2z",
                "lnpdraw_mass1_source",
                "lnpdraw_mass2_source_GIVEN_mass1_source",
                "lnpdraw_z",
                "lnpdraw_spin1_magnitude",
                "lnpdraw_spin2_magnitude",
                "lnpdraw_spin1_polar_angle",
                "lnpdraw_spin2_polar_angle",
                "v1_1ifo",
                "weights",
                "weights_1ifo",
                "name",
                "observed_phase_maximized_snr_net",
                "observed_snr_net",
                "optimal_snr_net",
                "semianalytic_observed_phase_maximized_snr_net",
            }
            keys = list(keys_of_interest.intersection(ff["events"].dtype.names))
            for substr in ["far", "ifar"]:
                keys += [
                    key
                    for key in ff["events"].dtype.names
                    if any(
                        [
                            key.startswith(f"{substr}_"),
                            key.endswith(f"_{substr}"),
                            f"_{substr}_" in key,
                        ]
                    )
                ]

            data = {key: np.array(ff["events"][key][()]) for key in keys}
            total_generated = int(ff.attrs["total_generated"][()])
            # the name applied to the analysis time changes between files, so we
            # loop over all plausible values and break once we find one
            for key in [
                "total_analysis_time",
                "analysis_time",
                "total_analysis_time_1ifo",
            ]:
                if key in ff.attrs:
                    analysis_time = ff.attrs[key][()] / 365.25 / 24 / 60 / 60
                    break
            else:
                raise AttributeError(
                    "Provided injection file does not provide analysis time"
                )
            if analysis_time == 0:
                analysis_time = 1 / 12
        else:
            raise KeyError(f"Unable to identify injections from {ff.keys()}")

        if "mass1_source" in data:
            mass_1_key = "mass1_source"
            mass_2_key = "mass2_source"
        else:
            mass_1_key = "mass_1_source"
            mass_2_key = "mass_2_source"
        if "redshift" in data:
            redshift_key = "redshift"
        else:
            redshift_key = "z"
        found_shape = data[mass_1_key][()].shape
        found = get_found_injections(data, found_shape, ifar_threshold, snr_threshold)
        n_found = sum(found)
        if n_found == 0:
            raise ValueError("No sensitivity injections pass threshold.")
        gwpop_data = dict(
            mass_1=xp.asarray(data[mass_1_key][()][found]),
            mass_ratio=xp.asarray(
                data[mass_2_key][()][found] / data[mass_1_key][()][found]
            ),
            redshift=xp.asarray(data[redshift_key][()][found]),
            total_generated=total_generated,
            analysis_time=analysis_time,
        )
        for ii in [1, 2]:
            gwpop_data[f"a_{ii}"] = (
                xp.asarray(
                    data.get(f"spin{ii}x", np.zeros(n_found))[()][found] ** 2
                    + data.get(f"spin{ii}y", np.zeros(n_found))[()][found] ** 2
                    + data[f"spin{ii}z"][()][found] ** 2
                )
                ** 0.5
            )
            gwpop_data[f"cos_tilt_{ii}"] = (
                xp.asarray(data[f"spin{ii}z"][()][found]) / gwpop_data[f"a_{ii}"]
            )
        if (
            "sampling_pdf" in data
        ):  # O1+O2+O3 mixture and endO3 injections (https://dcc.ligo.org/LIGO-T2100377, https://dcc.ligo.org/LIGO-T2100113)
            gwpop_data["prior"] = (
                xp.asarray(data["sampling_pdf"][()][found])
                * xp.asarray(data[mass_1_key][()][found])
                * (2 * np.pi * gwpop_data["a_1"] ** 2)
                * (2 * np.pi * gwpop_data["a_2"] ** 2)
            )
        elif (
            "lnpdraw_mass1_source_mass2_source_redshift_spin1x_spin1y_spin1z_spin2x_spin2y_spin2z"
            in data
        ):  # O1+O2+O3+O4a mixture (https://dcc.ligo.org/LIGO-T2400110)
            gwpop_data["prior"] = xp.exp(
                xp.asarray(
                    data[
                        "lnpdraw_mass1_source_mass2_source_redshift_spin1x_spin1y_spin1z_spin2x_spin2y_spin2z"
                    ][()][found]
                )
                + xp.log(xp.asarray(data[mass_1_key][()][found]))
                + xp.log(2 * np.pi * gwpop_data["a_1"] ** 2)
                + xp.log(2 * np.pi * gwpop_data["a_2"] ** 2)
            )
        else:  # O4a sensitivity injections (https://dcc.ligo.org/LIGO-T2400073)
            gwpop_data["prior"] = xp.exp(
                xp.sum(
                    [
                        xp.asarray(data[f"lnpdraw_{key}"][()][found])
                        for key in [
                            "mass1_source",
                            "mass2_source_GIVEN_mass1_source",
                            "z",
                            "spin1_magnitude",
                            "spin2_magnitude",
                            "spin1_polar_angle",
                            "spin2_polar_angle",
                        ]
                    ],
                    axis=0,
                )
            )
            gwpop_data["prior"] /= xp.sin(xp.arccos(gwpop_data["cos_tilt_1"]))
            gwpop_data["prior"] /= xp.sin(xp.arccos(gwpop_data["cos_tilt_2"]))
            gwpop_data["prior"] *= gwpop_data["mass_1"]

        weights = 1
        if "v1_1ifo" in vt_file:
            weights *= xp.asarray(data["weights_1ifo"][()][found])
        elif "weights" in data:
            weights *= xp.asarray(data["weights"][()][found])
        gwpop_data["prior"] /= weights
    return VTData(**gwpop_data)


def get_found_injections(data, shape, ifar_threshold=1, snr_threshold=11):
    found = np.zeros(shape, dtype=bool)
    has_ifar = any(["ifar" in key.lower() for key in data.keys()])

    far_keys = list(
        filter(
            lambda key: (
                key.lower().startswith("far_")
                or key.lower().endswith("_far")
                or "_far_" in key.lower()
            ),
            data,
        )
    )

    if not has_ifar and len(far_keys) > 0:
        for far_key in far_keys:
            data[far_key.replace("far", "ifar")] = 1 / data[far_key][()]
        has_ifar = True
    if ifar_threshold is None:
        ifar_threshold = 1e300
    if has_ifar:
        for key in data:
            if "ifar" in key.lower():
                found |= data[key][()] > ifar_threshold
            if "name" in data.keys():
                gwtc1 = (data["name"][()] == b"o1") | (data["name"][()] == b"o2")
                found |= gwtc1 & (data["optimal_snr_net"][()] > snr_threshold)
        if "semianalytic_observed_phase_maximized_snr_net" in data.keys():
            found |= (
                data["semianalytic_observed_phase_maximized_snr_net"][()]
                > snr_threshold
            )
        return found
    elif snr_threshold is not None:
        if "observed_phase_maximized_snr_net" in data.keys():
            found |= data["observed_phase_maximized_snr_net"][()] > snr_threshold
        elif "observed_snr_net" in data.keys():
            found |= data["observed_snr_net"][()] > snr_threshold
        return found
    else:
        raise ValueError("Cannot find keys to filter sensitivity injections.")


def apply_injection_prior(data, parameters):
    """
    We assume the injection prior is uniform in the detector frame primary
    mass and mass ratio.
    """
    from .data_collection import primary_mass_to_chirp_mass_jacobian, aligned_spin_prior

    if "mass_2" in parameters:
        data["mass_2"] = data["mass_1"] * data["mass_ratio"]
        data["prior"] /= data["mass_1"]
    if "chirp_mass" in parameters:
        jacobian = primary_mass_to_chirp_mass_jacobian(data)
        data["chirp_mass"] = data["mass_1"] / jacobian
        data["prior"] *= jacobian
    if "chi_eff" in parameters:
        data["chi_eff"] = chi_eff_from_components(
            a_1=data["a_1"],
            cos_tilt_1=data["cos_tilt_1"],
            a_2=data["a_2"],
            cos_tilt_2=data["cos_tilt_2"],
            mass_ratio=data["mass_ratio"],
        )
        if "chi_p" in parameters:
            data["chi_p"] = chi_p_from_components(
                a_1=data["a_1"],
                cos_tilt_1=data["cos_tilt_1"],
                a_2=data["a_2"],
                cos_tilt_2=data["cos_tilt_2"],
                mass_ratio=data["mass_ratio"],
            )
            amax = 1
            logger.info(
                f"Applying isotropic prior to chi_eff and chi_p, assuming injections with amax={amax}."
            )
            p_chi_iso = prior_chieff_chip_isotropic(
                data["chi_eff"], data["chi_p"], data["mass_ratio"], amax=amax
            )
            p_magnitude_costilt_iso = (1 / 2) ** 2 * (1 / amax) ** 2
            data["prior"] *= p_chi_iso / p_magnitude_costilt_iso
    if "chi_1" in parameters:
        data["chi_1"] = data["a_1"] * data["cos_tilt_1"]
        data["prior"] *= 2 * aligned_spin_prior(data["chi_1"])
    if "chi_2" in parameters:
        data["chi_2"] = data["a_2"] * data["cos_tilt_2"]
        data["prior"] *= 2 * aligned_spin_prior(data["chi_2"])
    if "mass_1_detector" in parameters:
        data["mass_1_detector"] = data["mass_1"] * (1 + data["redshift"])
        data["prior"] /= 1 + data["redshift"]
    if "mass_2_detector" in parameters:
        data["mass_2_detector"] = data["mass_1_detector"] * data["mass_ratio"]
        data["prior"] /= data["mass_1_detector"]
    if "chirp_mass_detector" in parameters:
        jacobian = primary_mass_to_chirp_mass_jacobian(data)
        try:
            data["chirp_mass_detector"] = data["mass_1_detector"] / jacobian
            data["prior"] *= jacobian
        except (KeyError, AttributeError, TypeError):
            data["chirp_mass_detector"] = (
                data["mass_1"] * (1 + data["redshift"]) / jacobian
            )
            data["prior"] *= jacobian / (1 + data["redshift"])
    if "luminosity_distance" in parameters:
        cosmo = get_cosmology("Planck15_LAL")

        data["luminosity_distance"] = cosmo.luminosity_distance(data["redshift"])
        data["prior"] /= cosmo.dDLdz(data["redshift"])


def dump_injection_data(args, save_filename=None):
    """
    Dump the injection data to a pickle file to :code:`{args.run_dir}/data/injections.pkl` for easier file transfer.

    This also makes sure the required parameters are present in the data and updates the prior accordingly.

    Parameters
    ----------
    args:
        Command-line arguments
    save_filename:
        The filename to save the data to. If None, it will save the data to args.run_dir/data/injections.pkl.
    """
    if "*" in args.vt_file:
        data = reduce(
            lambda x, y: x + y,
            [
                load_injection_data(
                    vt_file=get_path_or_local(vt_file),
                    ifar_threshold=args.vt_ifar_threshold,
                    snr_threshold=args.vt_snr_threshold,
                )
                for vt_file in glob.glob(args.vt_file)
            ],
        )
    else:
        data = load_injection_data(
            vt_file=get_path_or_local(args.vt_file),
            ifar_threshold=args.vt_ifar_threshold,
            snr_threshold=args.vt_snr_threshold,
        )
    apply_injection_prior(data, args.parameters)
    if save_filename is None:
        fname = f"{args.run_dir}/data/injections.pkl"
    else:
        fname = save_filename
    data_dict = data.to_dict()
    keys = args.parameters.copy() + ["prior", "total_generated", "analysis_time"]
    data_dict = {key: value for key, value in data_dict.items() if key in keys}
    with open(fname, "wb") as ff:
        dill.dump(data_dict, ff)
    logger.info(f"Written injection data to {fname}")


def injection_resampling_vt(vt_file, model, ifar_threshold=1, snr_threshold=11):

    if "*" in vt_file:
        vt_files = glob.glob(vt_file)
        data = sum(
            [
                load_injection_data(
                    vt_file=get_path_or_local(filename),
                    ifar_threshold=ifar_threshold,
                    snr_threshold=snr_threshold,
                )
                for filename in vt_files
            ]
        )
    else:
        data = load_injection_data(
            vt_file=get_path_or_local(vt_file),
            ifar_threshold=ifar_threshold,
            snr_threshold=snr_threshold,
        )

    return ResamplingVT(model=model, data=data.to_dict(), n_events=N_EVENTS)


def injection_resampling_vt_no_redshift(
    vt_file, model, ifar_threshold=1, snr_threshold=11
):

    data = load_injection_data(
        vt_file=vt_file, ifar_threshold=ifar_threshold, snr_threshold=snr_threshold
    )
    data["prior"] = data["mass_1"] ** (-2.35 + 1) * data["mass_ratio"] ** 2

    return ResamplingVT(model=model, data=data.to_dict(), n_events=N_EVENTS)


def create_injection_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "vt_file",
        type=str,
        help="File to load VT data from or a glob string matching multiple files to combine.",
    )

    parser.add_argument(
        "--vt-ifar-threshold",
        type=float,
        default=1,
        help="IFAR threshold for resampling injections",
    )
    parser.add_argument(
        "--vt-snr-threshold",
        type=float,
        default=11,
        help="SNR threshold for resampling injections. "
        "This is only used for O1/O2 injections",
    )
    parser.add_argument(
        "--vt-function",
        type=str,
        default="injection_resampling_vt",
        help="Function to generate selection function from.",
    )

    parser.add_argument(
        "--parameters",
        "-p",
        action="append",
        help=(
            "Parameters that are fit with the model. "
            "These are the parameters that will be extracted from the injections "
            "and should follow Bilby naming conventions with the exception that all masses "
            "are assumed to be in the source frame. Here is a list of parameters for which "
            "prior factors will be properly accounted. "
            "mass_1: source frame primary mass, mass_2: source frame secondary mass, "
            "chirp_mass: source frame chirp mass, mass_ratio: mass ratio, redshift: redshift, "
            "a_1: primary spin magnitude, a_2: secondary spin magnitude, cos_tilt_1: "
            "cosine primary spin tilt, cos_tilt_2: cosine secondary spin tilt, "
            "chi_1: aligned primary spin, chi_2: aligned secondary spin."
            "Any other parameters will be assumed to have a flat prior."
            "These parameters are also used to set the fiducial prior values. "
            "No redundancy checks are performed so users should be careful to not "
            "include unused parameters as that may have unintended consequences."
        ),
    )

    parser.add_argument(
        "--save_as",
        "-s",
        type=str,
        default="injections.pkl",
        help="name of pickle file to save the injections",
    )

    return parser


def read_injections():
    parser = create_injection_parser()
    args = parser.parse_args()
    save_file = args.save_as
    save_file = save_file.split(".pkl")[0] + ".pkl"
    dump_injection_data(args, save_file)



"""
Post-processing plotting functions. This makes spectra plots as well as plots
comparing fiducial and population-informed posteriors.

The module provides the `gwpopulation_pipe_plot` executable.
"""

#!/usr/bin/env python3
import os

import dill
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from bilby.core.result import plot_multiple, read_in_result
from bilby.core.utils import logger
from bilby.hyper.model import Model
from bilby_pipe.parser import StoreBoolean
from gwpopulation.backend import set_backend
from gwpopulation.conversions import convert_to_beta_parameters
from gwpopulation.models.mass import (
    BrokenPowerLawPeakSmoothedMassDistribution,
    BrokenPowerLawSmoothedMassDistribution,
    MultiPeakSmoothedMassDistribution,
    SinglePeakSmoothedMassDistribution,
    power_law_primary_mass_ratio,
)
from gwpopulation.models.redshift import MadauDickinsonRedshift, PowerLawRedshift
from gwpopulation.models.spin import (
    independent_spin_magnitude_beta,
    independent_spin_orientation_gaussian_isotropic,
)
from gwpopulation.utils import to_numpy, xp
from tqdm import tqdm

from .parser import create_parser as create_main_parser
from .data_analysis import _load_model

_new_matplotlib_settings = {
    "text.usetex": True,
    "font.serif": "Computer Modern Roman",
    "font.family": "Serif",
}

_old_matplotlib_settings = {
    key: matplotlib.rcParams[key] for key in _new_matplotlib_settings
}


def _set_matplotlib():
    for key in _new_matplotlib_settings:
        matplotlib.rcParams[key] = _new_matplotlib_settings[key]


def _unset_matplotlib():
    for key in _old_matplotlib_settings:
        matplotlib.rcParams[key] = _old_matplotlib_settings[key]


def _load_samples(filename):
    with open(filename, "rb") as ff:
        return dill.load(ff)


def _dump_samples(filename, data):
    with open(filename, "wb") as ff:
        dill.dump(data, file=ff)


def create_parser():
    parser = create_main_parser()
    parser.add_argument("--result-file", nargs="*")
    parser.add_argument("--labels", nargs="*", default=None)
    spectra = ["mass", "orientation", "magnitude", "redshift", "chi_eff_chi_p"]
    for key in spectra:
        parser.add_argument(
            f"--{key}",
            dest=key,
            action=StoreBoolean,
            default=True,
            help=f"Make a {key} spectrum plot?",
        )
    for key in ["mass", "magnitude", "tilt", "redshift", "chi_eff_chi_p"]:
        parser.add_argument(f"--{key}-model", default=None)
    parser.add_argument(
        "--corner",
        dest="corner",
        action=StoreBoolean,
        default=False,
        help="Make corner plots?",
    )
    parser.add_argument(
        "--lower-limit",
        type=float,
        default=2.5,
        help="Lower limit for confidence band.",
    )
    parser.add_argument(
        "--upper-limit",
        type=float,
        default=97.5,
        help="Upper limit for confidence band.",
    )
    parser.add_argument(
        "--samples",
        type=str,
        default=None,
        help="HDF5 file containing original and reampled posteriors.",
    )
    return parser


MAX_SAMPLES = 10000


def corner_plots(result, *args, **kwargs):
    """
    Make corner plots broken into distinct single-event parameters
    based on known hyperparemeter names, e.g., all parameters describing the
    mass spectrum.

    Parameters
    ----------
    result: `bilby.core.result.Result`
        The `Bilby` result object containing the posterior.
    args: unused
    kwargs: unused

    """
    mass_parameters = [
        "alpha",
        "mmax",
        "mmin",
        "mpp",
        "sigpp",
        "lam",
        "delta_m",
        "beta",
        "alpha_1",
        "alpha_2",
        "break_fraction",
    ]
    spin_magnitude_parameters = [
        "alpha_chi",
        "beta_chi",
        "mu_chi",
        "sigma_chi",
        "mu_chi_1",
        "mu_chi_2",
        "sigma_chi_1",
        "sigma_chi_2",
        "amax",
        "amax_1",
        "amax_2",
        "mu_chi_eff",
        "sigma_chi_eff",
        "mu_chi_p",
        "sigma_chi_p",
    ]
    spin_orientation_parameters = ["xi_spin", "sigma_1", "sigma_2", "sigma_spin"]
    chi_eff_chi_p_parameters = [
        "mu_chi_eff",
        "sigma_chi_eff",
        "mu_chi_p",
        "sigma_chi_p",
        "spin_covariance",
    ]
    redshift_parameters = ["lambda_z"]

    if len(result) > 1:
        for kind, pars in zip(
            ["mass", "spin_magnitude", "spin_orientation", "redshift", "chi_eff_chi_p"],
            [
                mass_parameters,
                spin_magnitude_parameters,
                spin_orientation_parameters,
                redshift_parameters,
                chi_eff_chi_p_parameters,
            ],
        ):
            plot_pars = [
                par
                for par in pars
                if all([par in res.search_parameter_keys for res in result])
            ]
            if len(plot_pars) == 0:
                continue
            logger.info(f"Making {kind} corner plot with: " + ", ".join(plot_pars))
            plot_multiple(
                results=result,
                parameters=plot_pars,
                quantiles=None,
                filename=f"{result[0].outdir}/combined_{kind}_corner.png",
            )
    else:
        result = result[0]
        for kind, pars in zip(
            ["mass", "spin_magnitude", "spin_orientation", "redshift", "chi_eff_chi_p"],
            [
                mass_parameters,
                spin_magnitude_parameters,
                spin_orientation_parameters,
                redshift_parameters,
                chi_eff_chi_p_parameters,
            ],
        ):
            plot_pars = [par for par in result.search_parameter_keys if par in pars]
            if len(plot_pars) == 0:
                continue
            logger.info(f"Making {kind} corner plot with: " + ", ".join(plot_pars))
            result.plot_corner(
                parameters=plot_pars,
                quantiles=None,
                filename=f"{result.outdir}/{result.label}_{kind}_corner.png",
            )


def mass_spectrum_plot(results, args, rate=False, observed=False, save=True):
    """
    Make the posterior mass spectrum plot given either one or multiple result
    objects.

    Parameters
    ----------
    results: [list, `bilby.core.result.Result`]
        `Bilby` result(s) to make the spectra from.
    args: ArgumentParser
        The arguments describing the model.
    rate: bool
        Whether to scale the spectra by the inferred rate.
    observed: bool
        Whether to correct for selection effects. Not currently implemented.
    save: bool
        Whether to save the result to file.

    Returns
    -------
    [dict, `matplotlib.pyplot.Figure`]
        If `save` the data behind the plot is returned as a dictionary.
        Else, the figure handle is returned.

    """
    mass_1 = xp.linspace(2, 100, 1000)
    mass_ratio = xp.linspace(0.1, 1, 500)
    mass_1_grid, mass_ratio_grid = xp.meshgrid(mass_1, mass_ratio)
    if not isinstance(results, list):
        result = [results]

    if observed:
        # vt_evaluator = load_vt(None)
        # vt_evaluator
        # TODO: make this usable...
        raise NotImplementedError(
            "Observation weighted spectrum plots not yet supported."
        )

    fig, axs = plt.subplots(2, 1, figsize=(12, 8))

    peak_1 = 0
    peak_2 = 0

    for result in results:
        filename = f"{result.outdir}/{result.label}_mass_data.h5"
        if os.path.isfile(filename):
            _data = _load_samples(filename)
            lines = _data["lines"]
            ppd = _data["ppd"]
            injected = _data.get("injected", None)
        else:
            data = dict(
                mass_1=mass_1_grid,
                mass_ratio=mass_ratio_grid,
                mass_2=mass_1_grid * mass_ratio_grid,
            )
            lines = dict(mass_1=list(), mass_ratio=list())
            ppd = xp.zeros_like(data["mass_1"])

            if len(result.posterior) > MAX_SAMPLES:
                samples = result.posterior.sample(MAX_SAMPLES)
            else:
                samples = result.posterior

            if model := result.meta_data["models"].get("mass", None):
                model = Model([_load_model(model, args)])
            elif args.mass_model is not None:
                model = Model([_load_model(args.mass_model, args)])
            elif "mpp_1" in samples.keys():
                model = Model([MultiPeakSmoothedMassDistribution()])
                if "delta_m" not in samples:
                    samples["delta_m"] = 0
            elif "alpha_2" in samples.keys():
                if "mpp" in samples.keys():
                    model = Model([BrokenPowerLawPeakSmoothedMassDistribution()])
                else:
                    model = Model([BrokenPowerLawSmoothedMassDistribution()])
                if "delta_m" not in samples:
                    samples["delta_m"] = 0
            elif "mpp" in samples.keys():
                model = Model([SinglePeakSmoothedMassDistribution()])
                if "delta_m" not in samples:
                    samples["delta_m"] = 0
            elif "alpha" in samples.keys():
                model = Model([power_law_primary_mass_ratio])

            for ii in tqdm(range(len(samples))):
                parameters = dict(samples.iloc[ii])
                model.parameters.update(parameters)
                prob = model.prob(data)
                if rate:
                    if "rate" not in parameters:
                        rate = False
                    else:
                        prob *= parameters["rate"]
                ppd += prob

                mass_1_prob = xp.trapz(prob, mass_ratio, axis=0)
                mass_ratio_prob = xp.trapz(prob, mass_1, axis=-1)

                lines["mass_1"].append(mass_1_prob)
                lines["mass_ratio"].append(mass_ratio_prob)
            for key in lines:
                lines[key] = np.vstack([to_numpy(line) for line in lines[key]])

            ppd /= len(samples)
            ppd = to_numpy(ppd)

            if result.injection_parameters is not None and not rate:
                model.parameters.update(result.injection_parameters)
                injected = to_numpy(model.prob(data))
            else:
                injected = None

            _dump_samples(filename, data=dict(lines=lines, ppd=ppd, injected=injected))

        mass_1 = to_numpy(mass_1)
        mass_ratio = to_numpy(mass_ratio)

        mass_1_ppd = np.trapz(ppd, mass_ratio, axis=0)
        mass_ratio_ppd = np.trapz(ppd, mass_1, axis=-1)

        label = " ".join(result.label.split("_")).title()

        axs[0].semilogy(mass_1, mass_1_ppd, label=label)
        axs[0].fill_between(
            mass_1,
            np.percentile(lines["mass_1"], args.lower_limit, axis=0),
            np.percentile(lines["mass_1"], args.upper_limit, axis=0),
            alpha=0.5,
        )
        _peak_1 = max(np.percentile(lines["mass_1"], args.upper_limit, axis=0))
        peak_1 = max(peak_1, _peak_1)
        axs[1].semilogy(mass_ratio, mass_ratio_ppd)
        axs[1].fill_between(
            mass_ratio,
            np.percentile(lines["mass_ratio"], args.lower_limit, axis=0),
            np.percentile(lines["mass_ratio"], args.upper_limit, axis=0),
            alpha=0.5,
        )
        _peak_2 = max(np.percentile(lines["mass_ratio"], args.upper_limit, axis=0))
        peak_2 = max(peak_2, _peak_2)

        if injected is not None:
            injected_mass_1 = np.trapz(injected, mass_ratio, axis=0)
            injected_mass_ratio = np.trapz(injected, mass_1, axis=-1)
            axs[0].plot(
                mass_1, injected_mass_1, color="k", label="True", linestyle="--"
            )
            axs[1].plot(mass_ratio, injected_mass_ratio, color="k", linestyle="--")

    axs[0].set_xlim(2, 100)
    axs[0].set_ylim(peak_1 / 100000, peak_1 * 1.1)
    axs[0].set_xlabel("$m_{1}$ [$M_{\\odot}$]")
    axs[0].legend(bbox_to_anchor=(0.5, 1.15), loc="upper center")
    if rate:
        ylabel = "$\\frac{d\\mathcal{R}}{dm_{1}}$ [Gpc$^{-3}$yr$^{-1}M_{\\odot}^{-1}$]"
    else:
        ylabel = "$p(m_{1})$ [$M_{\\odot}^{-1}$]"
    axs[0].set_ylabel(ylabel)

    axs[1].set_xlim(0.1, 1)
    axs[1].set_ylim(peak_2 / 10000, peak_2 * 1.1)
    axs[1].set_xlabel("$q$")
    if rate:
        ylabel = "$\\frac{d\\mathcal{R}}{dq}$ [Gpc$^{-3}$yr$^{-1}$]"
    else:
        ylabel = "$p(q)$"
    axs[1].set_ylabel(ylabel)

    if len(results) == 1:
        file_name = f"{result.outdir}/{result.label}_mass_spectrum.pdf"
    else:
        file_name = f"{result.outdir}/combined_mass_spectrum.pdf"
    plt.tight_layout()
    if save:
        plt.savefig(file_name, format="pdf", dpi=600, bbox_inches="tight")
        plt.close()
        return ppd
    else:
        return fig


def chi_eff_chi_p_plot(results, args, rate=False, observed=False, save=True):
    """
    Make the posterior chi_eff chi_p plot given either one or multiple result
    objects.

    Parameters
    ----------
    results: [list, `bilby.core.result.Result`]
        `Bilby` result(s) to make the spectra from.
    args: ArgumentParser
        The arguments describing the model.
    rate: bool
        Whether to scale the spectra by the inferred rate.
    observed: bool
        Whether to correct for selection effects. Not currently implemented.
    save: bool
        Whether to save the result to file.

    Returns
    -------
    [dict, `matplotlib.pyplot.Figure`]
        If `save` the data behind the plot is returned as a dictionary.
        Else, the figure handle is returned.

    """

    chi_eff = xp.linspace(-1, 1, 1000)
    chi_p = xp.linspace(0, 1, 500)
    chi_eff_grid, chi_p_grid = xp.meshgrid(chi_eff, chi_p)
    if not isinstance(results, list):
        result = [results]

    if observed:
        # vt_evaluator = load_vt(None)
        # vt_evaluator
        # TODO: make this usable...
        raise NotImplementedError(
            "Observation weighted spectrum plots not yet supported."
        )

    fig, axs = plt.subplots(2, 1, figsize=(12, 8))

    for result in results:
        filename = f"{result.outdir}/{result.label}_spin_data.h5"
        if os.path.isfile(filename):
            _data = _load_samples(filename)
            lines = _data["lines"]
            ppd = _data["ppd"]
            injected = _data.get("injected", None)
        else:
            data = dict(
                chi_eff=chi_eff_grid,
                chi_p=chi_p_grid,
            )
            lines = dict(chi_eff=list(), chi_p=list())
            ppd = xp.zeros_like(data["chi_eff"])

            if len(result.posterior) > MAX_SAMPLES:
                samples = result.posterior.sample(MAX_SAMPLES)
            else:
                samples = result.posterior

            try:
                try:
                    model = result.meta_data["models"]["spin"]
                except KeyError:
                    model = result.meta_data["models"]["effective_spin"]
            except KeyError:
                raise KeyError("No chi_eff chi_p spin model found in result.")
            model = Model([_load_model(model, args)])

            for ii in tqdm(range(len(samples))):
                parameters = dict(samples.iloc[ii])
                model.parameters.update(parameters)
                prob = model.prob(data)
                if rate:
                    if "rate" not in parameters:
                        rate = False
                    else:
                        prob *= parameters["rate"]
                ppd += prob

                chi_eff_prob = xp.trapz(prob, chi_p, axis=0)
                chi_p_prob = xp.trapz(prob, chi_eff, axis=-1)

                lines["chi_eff"].append(chi_eff_prob)
                lines["chi_p"].append(chi_p_prob)
            for key in lines:
                lines[key] = np.vstack([to_numpy(line) for line in lines[key]])

            ppd /= len(samples)
            ppd = to_numpy(ppd)

            if result.injection_parameters is not None and not rate:
                model.parameters.update(result.injection_parameters)
                injected = to_numpy(model.prob(data))
            else:
                injected = None

            _dump_samples(filename, data=dict(lines=lines, ppd=ppd, injected=injected))

        chi_eff = to_numpy(chi_eff)
        chi_p = to_numpy(chi_p)

        chi_eff_ppd = np.trapz(ppd, chi_p, axis=0)
        chi_p_ppd = np.trapz(ppd, chi_eff, axis=1)

        label = " ".join(result.label.split("_")).title()

        axs[0].plot(chi_eff, chi_eff_ppd, label=label)
        axs[0].fill_between(
            chi_eff,
            np.percentile(lines["chi_eff"], args.lower_limit, axis=0),
            np.percentile(lines["chi_eff"], args.upper_limit, axis=0),
            alpha=0.5,
        )

        axs[1].plot(chi_p, chi_p_ppd)
        axs[1].fill_between(
            chi_p,
            np.percentile(lines["chi_p"], args.lower_limit, axis=0),
            np.percentile(lines["chi_p"], args.upper_limit, axis=0),
            alpha=0.5,
        )

        if injected is not None:
            injected_chi_p = np.trapz(injected, chi_p, axis=0)
            injected_chi_eff = np.trapz(injected, chi_eff, axis=-1)
            axs[0].plot(
                chi_eff, injected_chi_eff, color="k", label="True", linestyle="--"
            )
            axs[1].plot(chi_p, injected_chi_p, color="k", linestyle="--")

    axs[0].set_xlim(-1, 1)
    axs[0].set_ylim(0)
    axs[0].set_xlabel("$\\chi_{eff}$")
    axs[0].legend(loc="upper center")
    if rate:
        ylabel = "$\\frac{d\\mathcal{R}}{d\\chi_{eff}}$ [Gpc$^{-3}$yr$^{-1}$]"
    else:
        ylabel = "$p(\\chi_{eff})$]"
    axs[0].set_ylabel(ylabel)

    axs[1].set_xlim(0.0, 1)
    axs[1].set_ylim(0)
    axs[1].set_xlabel("$\\chi_p$")
    if rate:
        ylabel = "$\\frac{d\\mathcal{R}}{d\\chi_p}$ [Gpc$^{-3}$yr$^{-1}$]"
    else:
        ylabel = "$p(\\chi_p)$"
    axs[1].set_ylabel(ylabel)

    if len(results) == 1:
        file_name = f"{result.outdir}/{result.label}_spin_spectrum.pdf"
    else:
        file_name = f"{result.outdir}/combined_spin_spectrum.pdf"
    plt.tight_layout()
    if save:
        plt.savefig(file_name, format="pdf", dpi=600, bbox_inches="tight")
        plt.close()
        return ppd
    else:
        return fig


def spin_magnitude_spectrum_plot(results, args, rate=False, save=True):
    """
    Make the posterior spin magnitude plot given either one or multiple result
    objects.

    Parameters
    ----------
    results: [list, `bilby.core.result.Result`]
        `Bilby` result(s) to make the spectra from.
    args: ArgumentParser
        The arguments describing the model.
    rate: bool
        Whether to scale the spectra by the inferred rate.
    save: bool
        Whether to save the result to file.

    Returns
    -------
    [dict, `matplotlib.pyplot.Figure`]
        If `save` the data behind the plot is returned as a dictionary.
        Else, the figure handle is returned.

    """
    mags = xp.linspace(0, 1, 1000)
    a_1, a_2 = xp.meshgrid(mags, mags)

    fig, axs = plt.subplots(1, 2, figsize=(12, 4))

    for result in results:
        filename = f"{result.outdir}/{result.label}_magnitude_data.h5"
        if os.path.isfile(filename):
            _data = _load_samples(filename)
            lines = _data["lines"]
            ppd = _data["ppd"]
            injected = _data.get("injected", None)
        else:
            if model := result.meta_data["models"].get("magnitude", None):
                model = Model([_load_model(model, args)])
            elif args.magnitude_model is not None:
                model = Model([_load_model(args.magnitude_model, args=args)])
            else:
                model = Model([independent_spin_magnitude_beta])
                for key in ["amax", "mu_chi", "sigma_chi", "alpha_chi", "beta_chi"]:
                    if key in result.posterior and f"{key}_1" not in result.posterior:
                        result.posterior[f"{key}_1"] = result.posterior[key]
                        result.posterior[f"{key}_2"] = result.posterior[key]
                        del result.posterior[key]
                result.posterior = convert_to_beta_parameters(result.posterior)[0]

            data = dict(a_1=a_1, a_2=a_2)
            lines = dict(a_1=list(), a_2=list())
            ppd = xp.zeros_like(data["a_1"])

            if len(result.posterior) > MAX_SAMPLES:
                samples = result.posterior.sample(MAX_SAMPLES)
            else:
                samples = result.posterior

            for ii in tqdm(range(len(samples))):
                parameters = dict(samples.iloc[ii])
                model.parameters.update(parameters)
                prob = model.prob(data)
                if rate:
                    if "rate" not in parameters:
                        rate = False
                    else:
                        prob *= parameters["rate"]
                ppd += prob

                a_1_prob = xp.trapz(prob, mags, axis=0)
                a_2_prob = xp.trapz(prob, mags, axis=-1)

                lines["a_1"].append(a_1_prob)
                lines["a_2"].append(a_2_prob)
            for key in lines:
                lines[key] = np.vstack([to_numpy(line) for line in lines[key]])

            ppd /= len(samples)
            ppd = to_numpy(ppd)

            if result.injection_parameters is not None and not rate:
                parameters = result.injection_parameters.copy()
                for key in ["amax", "mu_chi", "sigma_chi", "alpha_chi", "beta_chi"]:
                    if key in parameters and f"{key}_1" not in parameters:
                        parameters[f"{key}_1"] = parameters[key]
                        parameters[f"{key}_2"] = parameters[key]
                        del parameters[key]
                parameters = convert_to_beta_parameters(parameters)[0]
                model.parameters.update(parameters)
                injected = to_numpy(model.prob(data))
            else:
                injected = None

            _dump_samples(filename, data=dict(lines=lines, ppd=ppd))

        mags = to_numpy(mags)

        a_1_ppd = np.trapz(ppd, mags, axis=0)
        a_2_ppd = np.trapz(ppd, mags, axis=1)

        label = " ".join(result.label.split("_")).title()

        axs[0].plot(mags, a_1_ppd, label=label)
        axs[0].fill_between(
            mags,
            np.percentile(lines["a_1"], args.lower_limit, axis=0),
            np.percentile(lines["a_1"], args.upper_limit, axis=0),
            alpha=0.5,
        )
        axs[1].plot(mags, a_2_ppd)
        axs[1].fill_between(
            mags,
            np.percentile(lines["a_2"], args.lower_limit, axis=0),
            np.percentile(lines["a_2"], args.upper_limit, axis=0),
            alpha=0.5,
        )

        if injected is not None and not rate:
            a_1_injected = np.trapz(injected, mags, axis=0)
            a_2_injected = np.trapz(injected, mags, axis=1)
            axs[0].plot(mags, a_1_injected, color="k", label="True", linestyle="--")
            axs[1].plot(mags, a_2_injected, color="k", linestyle="--")

    for ii in [1, 2]:
        axs[ii - 1].set_xlim(0, 1)
        axs[ii - 1].set_ylim(0)
        axs[ii - 1].set_xlabel(f"$a_{ii}$")
        if rate:
            ylabel = f"$\\frac{{dN}}{{da_{ii}}}$"
        else:
            ylabel = f"$p(a_{ii})$"
        axs[ii - 1].set_ylabel(ylabel)
    axs[0].legend(bbox_to_anchor=(0.5, 1.15), loc="upper center")

    if len(results) == 1:
        file_name = f"{result.outdir}/{result.label}_magnitude_spectrum.pdf"
    else:
        file_name = f"{result.outdir}/comparison_magnitude_spectrum.pdf"
    if save:
        plt.savefig(file_name, format="pdf", dpi=600, bbox_inches="tight")
        plt.close()
        return ppd
    else:
        return fig


def spin_orientation_spectrum_plot(results, args, rate=False, save=True):
    """
    Make the posterior spin orientation plot given either one or multiple result
    objects.

    Parameters
    ----------
    results: [list, `bilby.core.result.Result`]
        `Bilby` result(s) to make the spectra from.
    args: ArgumentParser
        The arguments describing the model.
    rate: bool
        Whether to scale the spectra by the inferred rate.
    save: bool
        Whether to save the result to file.

    Returns
    -------
    [dict, `matplotlib.pyplot.Figure`]
        If `save` the data behind the plot is returned as a dictionary.
        Else, the figure handle is returned.

    """
    cos_tilts = xp.linspace(-1, 1, 1000)
    cos_tilt_1, cos_tilt_2 = xp.meshgrid(cos_tilts, cos_tilts)

    fig, axs = plt.subplots(1, 2, figsize=(12, 4))

    for result in results:
        filename = f"{result.outdir}/{result.label}_orientation_data.h5"
        if os.path.isfile(filename):
            _data = _load_samples(filename)
            lines = _data["lines"]
            ppd = _data["ppd"]
            injected = _data.get("injected", None)
        else:
            if model := result.meta_data["models"].get("tilt", None):
                model = Model([_load_model(model, args)])
            elif args.tilt_model is not None:
                model = Model([_load_model(args.tilt_model, args=args)])
            else:
                model = Model([independent_spin_orientation_gaussian_isotropic])
                if "sigma_1" not in result.posterior:
                    result.posterior["sigma_1"] = result.posterior["sigma_spin"]
                    result.posterior["sigma_2"] = result.posterior["sigma_spin"]

            data = dict(cos_tilt_1=cos_tilt_1, cos_tilt_2=cos_tilt_2)
            lines = dict(cos_tilt_1=list(), cos_tilt_2=list())
            ppd = xp.zeros_like(data["cos_tilt_1"])

            if len(result.posterior) > MAX_SAMPLES:
                samples = result.posterior.sample(MAX_SAMPLES)
            else:
                samples = result.posterior

            for ii in tqdm(range(len(samples))):
                parameters = dict(samples.iloc[ii])
                model.parameters.update(parameters)
                prob = model.prob(data)
                if rate:
                    if "rate" not in parameters:
                        rate = False
                    else:
                        prob *= parameters["rate"]
                ppd += prob

                cos_tilt_1_prob = xp.trapz(prob, cos_tilts, axis=0)
                cos_tilt_2_prob = xp.trapz(prob, cos_tilts, axis=1)

                lines["cos_tilt_1"].append(cos_tilt_1_prob)
                lines["cos_tilt_2"].append(cos_tilt_2_prob)
            for key in lines:
                lines[key] = np.vstack([to_numpy(line) for line in lines[key]])

            ppd /= len(samples)
            ppd = to_numpy(ppd)

            if result.injection_parameters is not None and not rate:
                parameters = result.injection_parameters
                if "sigma_1" not in parameters:
                    parameters["sigma_1"] = parameters["sigma_spin"]
                    parameters["sigma_2"] = parameters["sigma_spin"]
                model.parameters.update(parameters)
                injected = to_numpy(model.prob(data))
            else:
                injected = None

            _dump_samples(filename, data=dict(lines=lines, ppd=ppd))

        cos_tilts = to_numpy(cos_tilts)

        cos_tilt_1_ppd = np.trapz(ppd, cos_tilts, axis=0)
        cos_tilt_2_ppd = np.trapz(ppd, cos_tilts, axis=1)

        label = " ".join(result.label.split("_")).title()

        axs[0].plot(cos_tilts, cos_tilt_1_ppd, label=label)
        axs[0].fill_between(
            cos_tilts,
            np.percentile(lines["cos_tilt_1"], args.lower_limit, axis=0),
            np.percentile(lines["cos_tilt_1"], args.upper_limit, axis=0),
            alpha=0.5,
        )
        axs[1].plot(cos_tilts, cos_tilt_2_ppd)
        axs[1].fill_between(
            cos_tilts,
            np.percentile(lines["cos_tilt_2"], args.lower_limit, axis=0),
            np.percentile(lines["cos_tilt_2"], args.upper_limit, axis=0),
            alpha=0.5,
        )

        if injected is not None and not rate:
            t_1_injected = np.trapz(injected, cos_tilts, axis=0)
            t_2_injected = np.trapz(injected, cos_tilts, axis=1)
            axs[0].plot(
                cos_tilts, t_1_injected, color="k", label="True", linestyle="--"
            )
            axs[1].plot(cos_tilts, t_2_injected, color="k", linestyle="--")

    if rate:
        ylabel = "$\\frac{{d\\mathcal{{R}}}}{{d\\cos t_{}}}$ [Gpc$^{{-3}}$yr$^{{-1}}$]"
    else:
        ylabel = "$p(\\cos t_{})$"
    for ii in [1, 2]:
        axs[ii - 1].set_xlim(-1, 1)
        axs[ii - 1].set_ylim(0)
        axs[ii - 1].set_xlabel(f"$\\cos t_{ii}$")
        axs[ii - 1].set_ylabel(ylabel.format(str(ii)))
    axs[0].legend(bbox_to_anchor=(0.5, 1.15), loc="upper center")

    if len(results) == 1:
        file_name = f"{result.outdir}/{result.label}_orientation_spectrum.pdf"
    else:
        file_name = f"{result.outdir}/comparison_orientation_spectrum.pdf"
    if save:
        plt.savefig(file_name, format="pdf", dpi=600, bbox_inches="tight")
        plt.close()
        return ppd
    else:
        return fig


def redshift_spectrum_plot(results, args, rate=True, save=True):
    """
    Make the posterior redshift plot given either one or multiple result
    objects.

    Parameters
    ----------
    results: [list, `bilby.core.result.Result`]
        `Bilby` result(s) to make the spectra from.
    args: ArgumentParser
        The arguments describing the model.
    rate: bool
        Whether to scale the spectra by the inferred rate.
    save: bool
        Whether to save the result to file.

    Returns
    -------
    [dict, `matplotlib.pyplot.Figure`]
        If `save` the data behind the plot is returned as a dictionary.
        Else, the figure handle is returned.

    """
    fig, ax = plt.subplots(1, 1, figsize=(6, 4))

    for result in results:
        if model := result.meta_data["models"].get("redshift", None):
            model = _load_model(model, args)
        elif args.redshift_model is not None:
            model = _load_model(args.redshift_model, args=args)
        elif "lamb" in result.posterior:
            model = PowerLawRedshift(z_max=2.3)
        elif "gamma" in result.posterior:
            model = MadauDickinsonRedshift(z_max=args.upper_limit)
        else:
            raise KeyError("Cannot find redshift parameters.")

        redshifts = model.zs
        _np_redshifts = to_numpy(redshifts)

        filename = f"{result.outdir}/{result.label}_redshift_data.h5"
        if os.path.exists(filename):
            _data = _load_samples(filename)
            lines = _data["lines"]
            ppd = _data["ppd"]
        else:
            ppd = xp.zeros_like(redshifts)
            lines = dict(redshift=list())

            if len(result.posterior) > MAX_SAMPLES:
                samples = result.posterior.sample(MAX_SAMPLES)
            else:
                samples = result.posterior

            for ii in tqdm(range(len(samples))):
                parameters = dict(samples.iloc[ii])
                prob = model.psi_of_z(
                    redshift=redshifts,
                    **{key: parameters[key] for key in model.variable_names},
                )
                if rate:
                    if "surveyed_hypervolume" not in parameters:
                        rate = False
                    else:
                        prob *= parameters["rate"]
                ppd += prob

                lines["redshift"].append(to_numpy(prob))

            ppd /= len(samples)
            ppd = to_numpy(ppd)
            _dump_samples(filename, data=dict(lines=lines, ppd=ppd))

        label = " ".join(result.label.split("_")).title()

        ax.plot(_np_redshifts, ppd, label=label)
        ax.fill_between(
            _np_redshifts,
            np.percentile(lines["redshift"], args.lower_limit, axis=0),
            np.percentile(lines["redshift"], args.upper_limit, axis=0),
            alpha=0.5,
        )

    if rate:
        ylabel = "$\\mathcal{R}(z)$ [Gpc$^{-3}$yr$^{-1}$]"
    else:
        ylabel = "$\\frac{R(z)}{R(z=0)}$"
    ax.set_xlim(0, 2.3)
    ax.set_yscale("log")
    ax.set_xlabel("$z$")
    ax.set_ylabel(ylabel)
    ax.legend(bbox_to_anchor=(0.5, 1.15), loc="upper center")

    if len(results) == 1:
        file_name = f"{result.outdir}/{result.label}_redshift_spectrum.pdf"
    else:
        file_name = f"{result.outdir}/comparison_redshift_spectrum.pdf"
    if save:
        plt.savefig(file_name, format="pdf", dpi=600, bbox_inches="tight")
        plt.close()
        return ppd
    else:
        return fig


def reweighted_comparison(data, outdir="outdir"):
    """
    Plot a comparison of fiducial and population-informed single-event posteriors.

    The makes an interactive `html` figure with `plotly`.

    Parameters
    ----------
    data: dict
        Dictionary containing `original` and `reweighted` posteriors.
    outdir: str
        The output directory to save the file to.

    """
    import plotly.graph_objects as go
    from plotly.offline import plot
    from plotly.subplots import make_subplots

    parameter_names = list(data["original"].keys())
    plotting_parameters = list()
    for key in ["original", "reweighted"]:
        if "mass_1" in parameter_names:
            plotting_parameters.append("mass_1")
            if "mass_ratio" in parameter_names and "mass_2" not in parameter_names:
                data[key]["mass_2"] = data[key]["mass_1"] * data[key]["mass_ratio"]
            elif "mass_ratio" not in parameter_names and "mass_2" in parameter_names:
                data[key]["mass_ratio"] = data[key]["mass_2"] / data[key]["mass_1"]
        if all(
            [
                _key in data[key]
                for _key in ["a_1", "a_2", "cos_tilt_1", "cos_tilt_2", "mass_ratio"]
            ]
        ):
            data[key]["chi_eff"] = (
                data[key]["a_1"] * data[key]["cos_tilt_1"]
                + data[key]["mass_ratio"] * data[key]["a_2"] * data[key]["cos_tilt_2"]
            ) / (1 + data[key]["mass_ratio"])
            in_plane_1 = data[key]["a_1"] * (1 - data[key]["cos_tilt_1"] ** 2) ** 0.5
            in_plane_2 = data[key]["a_2"] * (1 - data[key]["cos_tilt_2"] ** 2) ** 0.5
            data[key]["chi_p"] = np.maximum(
                in_plane_1,
                data[key]["mass_ratio"]
                * (3 + 4 * data[key]["mass_ratio"])
                / (4 + 3 * data[key]["mass_ratio"])
                * in_plane_2,
            )
    if "chi_eff" in data["original"]:
        plotting_parameters.append("chi_eff")
        plotting_parameters.append("chi_p")
    plotting_parameters.append("mass_ratio")
    plotting_parameters.append("mass_2")
    if "names" not in data:
        names = [str(ii) for ii in range(data["original"][parameter_names[0]].shape[0])]
    else:
        names = data["names"]

    fig = make_subplots(
        rows=len(parameter_names),
        cols=1,
        row_heights=[1 / len(parameter_names)] * len(parameter_names),
    )

    fig.update_layout(
        template="plotly_white",
        font=dict(family="Computer Modern"),
        xaxis=dict(showgrid=False),
        yaxis=dict(showgrid=False),
    )

    kinds = list()
    parameters = list()
    legends = list()

    for jj, parameter in enumerate(plotting_parameters):
        for ii, event in enumerate(names):
            fig.add_trace(
                go.Histogram(
                    x=to_numpy(data["original"][parameter][ii])[:1000],
                    histnorm="probability density",
                    marker_color="Blue",
                    legendgroup=event,
                    showlegend=False,
                    name=event,
                    hovertext=event,
                ),
                row=jj + 1,
                col=1,
            )
            kinds.append("original")
            parameters.append(parameter)
            if jj == 0:
                legends.append(True)
            else:
                legends.append(False)
            fig.add_trace(
                go.Histogram(
                    x=to_numpy(data["reweighted"][parameter][ii])[:1000],
                    histnorm="probability density",
                    marker_color="Red",
                    legendgroup=event,
                    showlegend=True and jj == 0,
                    name=event,
                    hovertext=event,
                ),
                row=jj + 1,
                col=1,
            )
            kinds.append("resampled")
            parameters.append(parameter)
            if jj == 0:
                legends.append(True)
            else:
                legends.append(False)
        fig.update_xaxes(title_text=" ".join(parameter.split("_")), row=jj + 1, col=1)

    fig.update_layout(barmode="overlay")
    fig.update_traces(opacity=0.3)

    updatemenus = [
        go.layout.Updatemenu(
            type="buttons",
            direction="right",
            active=0,
            x=0.57,
            y=1.2,
            buttons=list(
                [
                    dict(
                        label="All",
                        method="update",
                        args=[
                            {
                                "visible": [True] * len(kinds),
                                "showlegend": [
                                    kind == "original" and legend
                                    for kind, legend in zip(kinds, legends)
                                ],
                            }
                        ],
                    ),
                    dict(
                        label="Original",
                        method="update",
                        args=[
                            {
                                "visible": [kind == "original" for kind in kinds],
                                "showlegend": [
                                    kind == "original" and legend
                                    for kind, legend in zip(kinds, legends)
                                ],
                            }
                        ],
                    ),
                    dict(
                        label="Resampled",
                        method="update",
                        args=[
                            {
                                "visible": [kind == "resampled" for kind in kinds],
                                "showlegend": [
                                    kind == "original" and legend
                                    for kind, legend in zip(kinds, legends)
                                ],
                            }
                        ],
                    ),
                ]
            ),
        )
    ]
    fig.update_layout(updatemenus=updatemenus)
    fig.update_layout(height=250 * len(plotting_parameters), width=1000)
    filename = f"{outdir}/{data['label']}_samples.html"
    plot(fig, filename=filename, include_mathjax="cdn", auto_open=False)


PLOT_MAP = dict(
    mass=mass_spectrum_plot,
    magnitude=spin_magnitude_spectrum_plot,
    orientation=spin_orientation_spectrum_plot,
    redshift=redshift_spectrum_plot,
    corner=corner_plots,
    chi_eff_chi_p=chi_eff_chi_p_plot,
)


def _safe_plot(result, args, key):
    _set_matplotlib()
    try:
        PLOT_MAP[key](result, args, rate=True)
    except Exception as e:
        if "TeX" in e.args[0]:
            logger.warning(
                f"Failed to create plot for {key} due to missing TeX. " "Disabling TeX."
            )
            _unset_matplotlib()
            PLOT_MAP[key](result, args, rate=True)
        else:
            logger.warning(f"Failed to create {key} plot with message {e}")
    _unset_matplotlib()


def main():
    parser = create_parser()
    args, _ = parser.parse_known_args()
    set_backend(args.backend)
    results = [read_in_result(res) for res in args.result_file]
    if args.labels is not None:
        for label, result in zip(args.labels, results):
            result.label = label
    if args.run_dir is not None:
        for result in results:
            result.outdir = os.path.join(args.run_dir, "result")
    for key in [
        "mass",
        "magnitude",
        "orientation",
        "redshift",
        "corner",
        "chi_eff_chi_p",
    ]:
        if getattr(args, key):
            logger.info(f"Making {key} plot")
            _safe_plot(results, args, key)
    if args.samples is not None and os.path.exists(args.samples):
        samples = _load_samples(args.samples)
        reweighted_comparison(data=samples, outdir=os.path.join(args.run_dir, "result"))
    elif args.samples is not None:
        logger.info(
            f"Cannot find samples file {args.samples}. \
            Skipping reweighted comparison plot."
        )


if __name__ == "__main__":
    main()



import bilby
import gwpopulation
import wcosmo
import numpy as np
from bilby_pipe.bilbyargparser import BilbyArgParser
from bilby_pipe.parser import StoreBoolean
from bilby_pipe.utils import noneint, nonestr


def create_parser():
    from ._version import __version__

    parser = BilbyArgParser(
        usage=__doc__, ignore_unknown_config_file_keys=False, allow_abbrev=False
    )
    parser.add("ini", type=str, is_config_file=True, help="Configuration ini file")
    parser.add("-v", "--verbose", action="store_true", help="Verbose output")
    parser.add(
        "--version",
        action="version",
        version="%(prog)s={version}\nbilby={bilby_version}\ngwpopulation={gwpopulation_version}".format(
            version=__version__,
            bilby_version=bilby.__version__,
            gwpopulation_version=gwpopulation.__version__,
        ),
    )

    base_parser = parser.add_argument_group(
        title="Generic arguments", description="Generic arguments"
    )

    base_parser.add_argument(
        "--run-dir",
        type=str,
        default="outdir",
        help="Output directory for posterior samples",
    )
    base_parser.add_argument(
        "--log-dir",
        type=str,
        default="logs",
        help="Output directory for writing log files",
    )
    base_parser.add_argument("--label", type=str, default="label", help="Run label")
    base_parser.add_argument("--user", type=str, help="User name", default=None)
    base_parser.add_argument(
        "--vt-file",
        type=nonestr,
        default=None,
        help="File to load VT data from or a glob string matching multiple files to combine.",
    )
    base_parser.add_argument(
        "--vt-ifar-threshold",
        type=float,
        default=1,
        help="IFAR threshold for resampling injections",
    )
    base_parser.add_argument(
        "--vt-snr-threshold",
        type=float,
        default=11,
        help="IFAR threshold for resampling injections. "
        "This is only used for O1/O2 injections",
    )
    base_parser.add_argument(
        "--vt-function",
        type=str,
        default="injection_resampling_vt",
        help="Function to generate selection function from.",
    )
    base_parser.add_argument(
        "--prior-file",
        type=str,
        help="Prior file containing priors for all considered parameters",
    )
    base_parser.add_argument(
        "--request-gpu",
        default=0,
        help="Whether to request a GPU for the relevant jobs.",
    )
    base_parser.add_argument(
        "--require-gpus",
        default='DeviceName=="GeForce GTX 1050 Ti"',
        type=str,
        help="The GPU requirements to pass for HTCondor.",
    )
    base_parser.add_argument(
        "--backend",
        default="jax",
        choices=gwpopulation.backend.SUPPORTED_BACKENDS,
        type=str,
        help="The backend to use for the analysis, default is jax",
    )
    base_parser.add_argument(
        "--cosmo",
        action=StoreBoolean,
        default=False,
        help="Whether to fit cosmological parameters.",
    )
    base_parser.add_argument(
        "--cosmology",
        type=str,
        default="Planck15_LAL",
        help=(
            "Cosmology to use for the analysis, this should be one of "
            f"{', '.join(wcosmo.available.keys())}, Planck15_LAL. Some of these are "
            "fixed pre-defined cosmologies while others are parameterized "
            "cosmologies. If a parameterized cosmology is used the parameters relevant"
            " parameters should be included in the prior specification."
        ),
    )
    base_parser.add_argument(
        "--container",
        type=nonestr,
        default=None,
        help="The path to the singularity image to use",
    )
    base_parser.add_argument(
        "--conda-env", type=nonestr, default=None, help="The conda environment to use"
    )
    base_parser.add_argument(
        "--pool",
        type=str,
        choices=["osg", "local"],
        default="osg",
        help="Which HTCondor pool to submit to, if osg, the local pool is also allowed",
    )

    model_parser = parser.add_argument_group(
        title="Analysis models", description="Analysis models"
    )
    model_parser.add_argument(
        "--all-models", type=str, help="All models to use, formatted as a json string"
    )
    model_parser.add_argument(
        "--source-files",
        action="append",
        help=(
            "Files containing source models to use for user-defined models. "
            "These files are transferred to the execute node when using the "
            "HTCondor file transfer mechanism. If the job is being run "
            "locally the file should be in the users PYTHONPATH."
        ),
    )

    collection_parser = parser.add_argument_group(
        title="Data collection arguments", description="Data collection arguments"
    )
    collection_parser.add_argument(
        "--existing-data-directory",
        type=str,
        default="/fail",
        help="Directory containing existing data",
    )
    collection_parser.add_argument(
        "--parameters",
        action="append",
        help=(
            "Parameters that are fit with the model. "
            "These are the parameters that will be extracted from the posterior samples "
            "and should follow Bilby naming conventions with the exception that all masses "
            "are assumed to be in the source frame. Here is a list of parameters for which "
            "prior factors will be properly accounted. "
            "mass_1: source frame primary mass, mass_2: source frame secondary mass, "
            "mass_1_detector: detector frame primary mass, mass_2_detector: detector frame secondary mass, "
            "chirp_mass: source frame chirp mass, chirp_mass_detector: detector frame chirp mass,"
            "mass_ratio: mass ratio, redshift: redshift, luminosity_distance: luminosity distance,"
            "a_1: primary spin magnitude, a_2: secondary spin magnitude, cos_tilt_1: "
            "cosine primary spin tilt, cos_tilt_2: cosine secondary spin tilt, "
            "chi_1: aligned primary spin, chi_2: aligned secondary spin."
            "Any other parameters will be assumed to have a flat prior."
            "These parameters are also used to set the fiducial prior values. "
            "No redundancy checks are performed so users should be careful to not "
            "include unused parameters as that may have unintended consequences."
        ),
    )
    collection_parser.add_argument(
        "--ignore", action="append", help="Events to ignore."
    )
    collection_parser.add_argument(
        "--sample-regex", type=str, help="Pattern to match for sample files"
    )
    collection_parser.add_argument(
        "--preferred-labels",
        action="append",
        help="Run labels to search for in sample files",
    )
    collection_parser.add_argument(
        "--plot",
        default=True,
        action=StoreBoolean,
        help="Whether to generate diagnostic plots",
    )
    collection_parser.add_argument(
        "--n-simulations", type=noneint, help="Number of posteriors to simulate"
    )
    collection_parser.add_argument(
        "--samples-per-posterior",
        type=int,
        default=int(1e6),
        help="Number of samples per posterior. If larger than the number of samples in the shortest posterior dataset, will ignore this input.",
    )
    collection_parser.add_argument(
        "--collection-seed",
        type=noneint,
        help="Seed for the downsampling of the posteriors for each event. For reproducibility.",
    )
    collection_parser.add_argument(
        "--data-label", default="posteriors", help="Label for data product."
    )
    collection_parser.add_argument(
        "--distance-prior",
        default="comoving",
        type=str,
        help=(
            "Distance prior format, e.g., euclidean, comoving. 'euclidean' assumes the distance prior goes "
            "like D^2. 'comoving' assumes sources are uniformly distributed in the comoving frame using "
            "the Planck15_LAL cosmology. Can be in the format of a dict with the same keys as the sample-regex"
        ),
    )

    collection_parser.add_argument(
        "--mass-prior",
        default="flat-detector",
        type=str,
        help=(
            "Mass prior used during the initial sampling, must match one of the following options. "
            "\n 'flat-detector': Flat in detector frame primary and secondary masses. "
            "\n 'chirp-mass': Flat in detector frame chirp mass and mass ratio. "
            "\n 'flat-detector-components': Flat in detector frame primary and secondary masses. "
            "This is the default for LVK samples and the same as the deprecated 'flat-detector' option. "
            "\n 'flat-source-components': Flat in source frame primary and secondary masses. "
            "\n 'flat-detector-chirp-mass-ratio': Flat in detector frame chirp mass and mass ratio. "
            "This is the same as the deprecated 'chirp-mass' option. Can be in the format of a dict with the same keys as the sample-regex"
        ),
    )
    collection_parser.add_argument(
        "--spin-prior",
        default="component",
        type=str,
        help=(
            "Spin prior, the only supported spin prior assumes the spins are isotropically distributed "
            "with a flat prior on the magnitude. Can be in the format of a dict with the same keys as the sample-regex."
        ),
    )

    analysis_parser = parser.add_argument_group(
        title="Arguments describing analysis jobs", description="Analysis arguments"
    )
    analysis_parser.add_argument(
        "--max-redshift",
        default=2.3,
        type=float,
        help="The maximum redshift considered, this should match the injections.",
    )
    analysis_parser.add_argument(
        "--minimum-mass",
        default=2,
        type=float,
        help="The minimum mass considered, this should match the injections "
        "and is important for smoothed mass models.",
    )
    analysis_parser.add_argument(
        "--maximum-mass",
        default=100,
        type=float,
        help="The maximum mass considered, this should match the injections "
        "and is important for smoothed mass models.",
    )
    analysis_parser.add_argument(
        "--sampler",
        default="dynesty",
        type=str,
        help="The sampler to use, the default is dynesty",
    )
    analysis_parser.add_argument(
        "--sampler-kwargs",
        type=str,
        default="Default",
        help=(
            "Dictionary of sampler-kwargs to pass in, e.g., {nlive: 1000} OR "
            "pass pre-defined set of sampler-kwargs {Default, FastTest}"
        ),
    )
    analysis_parser.add_argument(
        "--vt-parameters",
        action="append",
        help=(
            "Which parameters to include in the VT estimate, should be some "
            "combination of mass, redshift, spin parameters, see the '--parameters' "
            "option for more details."
        ),
    )
    analysis_parser.add_argument(
        "--enforce-minimum-neffective-per-event",
        action=StoreBoolean,
        default=True,
        help=(
            "Require that all Monte Carlo integrals for the single event "
            "marignalizaed likleihoods have at least as many effective samples"
            " as the number of events."
        ),
    )

    analysis_parser.add_argument(
        "--maximum-uncertainty",
        type=float,
        default=np.inf,
        help=(
            "Require that all log likelihood evaluations have an uncertainty (standard deviation) less than"
            " this value. It is not recommended to use with the '--enforce-minimum-neffective-per-event' option."
            "See Talbot and Golomb (2023) arxiv:2304.06138"
        ),
    )

    injection_parser = parser.add_argument_group(
        title="Arguments describing injections", description="Injection arguments"
    )
    injection_parser.add_argument(
        "--injection-file",
        default=None,
        type=nonestr,
        help="JSON file containing population parameters, should be pandas readable.",
    )
    injection_parser.add_argument(
        "--injection-index", type=noneint, help="Index in injection file to use."
    )
    injection_parser.add_argument(
        "--sample-from-prior",
        action=StoreBoolean,
        help="Simulate posteriors from prior.",
    )

    post_parser = parser.add_argument_group(
        title="Post processing arguments", description="Post arguments"
    )
    post_parser.add_argument(
        "--post-plots",
        action=StoreBoolean,
        default=True,
        help="Whether to make post-processing plots.",
    )
    post_parser.add_argument(
        "--make-summary",
        action=StoreBoolean,
        default=True,
        help="Whether to make a summary page.",
    )
    post_parser.add_argument(
        "--n-post-samples",
        default=5000,
        type=int,
        help="Number of samples to use in the common format script",
    )

    return parser
